@@ -8,8 +8,10 @@ MpPipe = multiprocessing.Pipe
 MpQueue = multiprocessing.Queue
 MpProcess = multiprocessing.Process
 ipdb_nl_async = True
+nlm_generator = False
 
 commit_barrier = 0
+gc_timeout = 60
 
 # save uname() on startup time: it is not so
 # highly possible that the kernel will be
