@@ -28,24 +28,30 @@ from pyroute2.ipdb.exceptions import CommitException
 from pyroute2.ipdb.exceptions import PartialCommitException
 
 
-groups = rtnl.RTNLGRP_LINK |\
-    rtnl.RTNLGRP_NEIGH |\
-    rtnl.RTNLGRP_IPV4_IFADDR |\
-    rtnl.RTNLGRP_IPV6_IFADDR
+supported_kinds = ('bridge',
+                   'bond',
+                   'tuntap',
+                   'vxlan',
+                   'gre',
+                   'gretap',
+                   'ip6gre',
+                   'ip6gretap',
+                   'macvlan',
+                   'macvtap',
+                   'ipvlan',
+                   'vrf')
+
+groups = rtnl.RTMGRP_LINK |\
+    rtnl.RTMGRP_NEIGH |\
+    rtnl.RTMGRP_IPV4_IFADDR |\
+    rtnl.RTMGRP_IPV6_IFADDR
 
 
 def _get_data_fields():
+    global supported_kinds
+
     ret = []
-    for data in ('bridge',
-                 'bond',
-                 'tuntap',
-                 'vxlan',
-                 'gre',
-                 'ip6gre',
-                 'macvlan',
-                 'macvtap',
-                 'ipvlan',
-                 'vrf'):
+    for data in supported_kinds:
         msg = ifinfmsg.ifinfo.data_map[data]
         if getattr(msg, 'prefix', None) is not None:
             ret += [msg.nla2name(i[0]) for i in msg.nla_map]
@@ -98,7 +104,7 @@ class Interface(Transactional):
                        'net_ns_fd',
                        'net_ns_pid']
     _fields = [ifinfmsg.nla2name(i[0]) for i in ifinfmsg.nla_map]
-    for name in ('bridge_data', 'bridge_slave_data'):
+    for name in ('bridge_slave_data', ):
         data = getattr(ifinfmsg.ifinfo, name)
         _fields.extend([ifinfmsg.nla2name(i[0]) for i in data.nla_map])
     _fields.append('index')
@@ -260,6 +266,8 @@ class Interface(Transactional):
         This call always bypasses open transactions, loading
         changes directly into the interface data.
         '''
+        global supported_kinds
+
         with self._direct_state:
             if self['ipdb_scope'] == 'locked':
                 # do not touch locked interfaces
@@ -302,9 +310,7 @@ class Interface(Transactional):
                         self['vlan_flags'] = data\
                             .get_attr('IFLA_VLAN_FLAGS', {})\
                             .get('flags', 0)
-                    if kind in ('vxlan', 'macvlan', 'macvtap', 'gre',
-                                'gretap', 'ipvlan', 'bridge',
-                                'ip6gre', 'ip6gretap'):
+                    if kind in supported_kinds:
                         data = linkinfo.get_attr('IFLA_INFO_DATA') or {}
                         for nla in data.get('attrs', []):
                             norm = ifinfmsg.nla2name(nla[0])
@@ -353,26 +359,33 @@ class Interface(Transactional):
         '''
         Add IP address to an interface
 
-        Keyword arguments:
+        Address formats:
 
-        * mask
-        * broadcast
-        * anycast
-        * scope
+            with ipdb.interfaces.eth0 as i:
+                i.add_ip('192.168.0.1', 24)
+                i.add_ip('192.168.0.2/24')
+                i.add_ip('192.168.0.3/255.255.255.0')
+                i.add_ip('192.168.0.4/24',
+                         broadcast='192.168.0.255',
+                         scope=254)
         '''
+        family = 0
         # split mask
         if mask is None:
             ip, mask = ip.split('/')
-            if mask.find('.') > -1:
-                mask = dqn2int(mask)
-            else:
-                mask = int(mask, 0)
-        elif isinstance(mask, basestring):
-            mask = dqn2int(mask)
 
-        # normalize the address
         if ip.find(':') > -1:
+            family = AF_INET6
+            # normalize IPv6 format
             ip = inet_ntop(AF_INET6, inet_pton(AF_INET6, ip))
+        else:
+            family = AF_INET
+
+        if isinstance(mask, basestring):
+            try:
+                mask = int(mask, 0)
+            except:
+                mask = dqn2int(mask, family)
 
         # if it is a transaction or an interface update, apply the change
         self['ipaddr'].unlink((ip, mask))
@@ -404,14 +417,14 @@ class Interface(Transactional):
             self['ipaddr'].remove((ip, mask))
 
     @with_transaction
-    def add_vlan(self, vlan):
+    def add_vlan(self, vlan, flags=None):
         if isinstance(vlan, dict):
             vid = vlan['vid']
         else:
             vid = vlan
             vlan = {'vid': vlan, 'flags': 0}
         self['vlans'].unlink(vid)
-        self['vlans'].add(vid, raw=vlan)
+        self['vlans'].add(vid, raw=(vlan, flags))
 
     @with_transaction
     def del_vlan(self, vlan):
@@ -516,19 +529,22 @@ class Interface(Transactional):
         error = None
         added = None
         removed = None
-        drop = True
+        drop = self.ipdb.txdrop
+        notx = True
+
         init = None
         debug = {'traceback': None,
                  'transaction': None,
                  'next_stage': None}
 
+        if tid or transaction:
+            notx = False
+
         if tid:
             transaction = self.global_tx[tid]
         else:
-            if transaction:
-                drop = False
-            else:
-                transaction = self.current_tx
+            transaction = transaction or self.current_tx
+
         if transaction.partial:
             transaction.errors = []
 
@@ -539,7 +555,7 @@ class Interface(Transactional):
                 # a special case: transition "create" -> "remove"
                 if transaction['ipdb_scope'] == 'remove' and \
                         self['ipdb_scope'] == 'create':
-                    self.ipdb.interfaces._detach(self['ifname'], 0)
+                    self.invalidate()
                     return self
 
                 newif = True
@@ -562,7 +578,9 @@ class Interface(Transactional):
                     init = self.pick()
                     try:
                         request = {key: transaction[key] for key in
-                                   filter(lambda x: x[:3] != 'br_',
+                                   filter(lambda x: x[:5] != 'bond_' and
+                                          x[:7] != 'brport_' and
+                                          x[:3] != 'br_',
                                           transaction)}
                         for key in ('net_ns_fd', 'net_ns_pid'):
                             if key in request:
@@ -608,12 +626,13 @@ class Interface(Transactional):
                         # broadcast will be sent, and the object is unmodified.
                         # After the exception forwarding, the object is ready
                         # to repeat the commit() call.
+                        if drop and notx:
+                            self.drop(transaction.uid)
                         raise
 
         if transaction['ipdb_scope'] == 'create' and commit_phase > 1:
             if self['index']:
-                wd = self.ipdb.watchdog(action='RTM_DELLINK',
-                                        ifname=self['ifname'])
+                wd = self.ipdb.watchdog('RTM_DELLINK', ifname=self['ifname'])
                 with self._direct_state:
                     self['ipdb_scope'] = 'locked'
                 self.nl.link('delete', index=self['index'])
@@ -625,6 +644,8 @@ class Interface(Transactional):
             # Here we come only if a new interface is created
             #
             if commit_phase == 1 and not self.wait_target('ipdb_scope'):
+                if drop and notx:
+                    self.drop(transaction.uid)
                 self.invalidate()
                 if isinstance(newif, Exception):
                     raise newif
@@ -677,11 +698,11 @@ class Interface(Transactional):
         resolve_ports(transaction,
                       transaction._delay_add_port,
                       transaction.add_port,
-                      self, drop)
+                      self, drop and notx)
         resolve_ports(transaction,
                       transaction._delay_del_port,
                       transaction.del_port,
-                      self, drop)
+                      self, drop and notx)
 
         try:
             removed, added = snapshot // transaction
@@ -693,23 +714,27 @@ class Interface(Transactional):
             # Port vlans
             if removed['vlans'] or added['vlans']:
 
-                if added['vlans']:
-                    transaction['vlans'].add(1)
                 self['vlans'].set_target(transaction['vlans'])
 
                 for i in removed['vlans']:
-                    if i != 1:
-                        # remove vlan from the port
-                        run(nl.vlan_filter, 'del',
-                            index=self['index'],
-                            vlan_info=self['vlans'][i])
+                    # remove vlan from the port
+                    run(nl.vlan_filter, 'del',
+                        index=self['index'],
+                        vlan_info=self['vlans'][i][0])
 
                 for i in added['vlans']:
-                    if i != 1:
-                        # add vlan to the port
-                        run(nl.vlan_filter, 'add',
-                            index=self['index'],
-                            vlan_info=transaction['vlans'][i])
+                    # add vlan to the port
+                    vinfo = transaction['vlans'][i][0]
+                    flags = transaction['vlans'][i][1]
+                    req = {'index': self['index'],
+                           'vlan_info': vinfo}
+                    if flags == 'self':
+                        req['vlan_flags'] = flags
+                        # this request will NOT give echo,
+                        # so bypass the check
+                        with self._direct_state:
+                            self.add_vlan(vinfo['vid'])
+                    run(nl.vlan_filter, 'add', **req)
 
                 self['vlans'].target.wait(SYNC_TIMEOUT)
                 if not self['vlans'].target.is_set():
@@ -741,7 +766,10 @@ class Interface(Transactional):
                         transaction.errors.append(KeyError(i))
 
                 self['ports'].target.wait(SYNC_TIMEOUT)
-                if not self['ports'].target.is_set():
+                if self['ports'].target.is_set():
+                    for msg in self.nl.get_vlans(index=self['index']):
+                        self.load_netlink(msg)
+                else:
                     raise CommitException('ports target is not set')
 
                 # 1. wait for proper targets on ports
@@ -788,6 +816,8 @@ class Interface(Transactional):
                     elif key[:7] == 'brport_':
                         prequest[key[7:]] = added[key]
                     else:
+                        if key == 'address' and added[key] is not None:
+                            self[key] = added[key].lower()
                         request[key] = added[key]
             # FIXME: flush the interface type so the next two conditions
             # will work correctly
@@ -965,8 +995,7 @@ class Interface(Transactional):
             # 8<---------------------------------------------
             # Interface removal
             if added.get('ipdb_scope') in ('shadow', 'remove'):
-                wd = self.ipdb.watchdog(action='RTM_DELLINK',
-                                        ifname=self['ifname'])
+                wd = self.ipdb.watchdog('RTM_DELLINK', ifname=self['ifname'])
                 with self._direct_state:
                     self['ipdb_scope'] = 'locked'
                 self.nl.link('delete', index=self['index'])
@@ -982,7 +1011,7 @@ class Interface(Transactional):
                 if added.get('ipdb_scope') == 'remove':
                     self.ipdb.interfaces._detach(None, self['index'], None)
 
-                if drop:
+                if notx:
                     self.drop(transaction.uid)
 
                 return self
@@ -992,6 +1021,9 @@ class Interface(Transactional):
             if commit_phase == 1:
                 self.ipdb.ensure('run')
 
+            # so far all's ok
+            drop = True
+
         except Exception as e:
             error = e
             # log the error environment
@@ -1032,8 +1064,8 @@ class Interface(Transactional):
         if transaction.partial and transaction.errors:
             error = PartialCommitException('partial commit error')
 
-        # if it is not a rollback turn
-        if drop and commit_phase == 1:
+        # drop only if required
+        if drop and notx:
             # drop last transaction in any case
             self.drop(transaction.uid)
 
@@ -1157,7 +1189,7 @@ class InterfacesDict(Dotkeys):
                     if key[-4:] == 'link':
                         if isinstance(kwarg[key], Interface):
                             kwarg[key] = kwarg[key].get('index') or \
-                                    kwarg[key].get('ifname')
+                                kwarg[key].get('ifname')
                         if not isinstance(kwarg[key], int):
                             device._deferred_link = (key, kwarg[key])
                 device._mode = self.ipdb.mode
@@ -1173,7 +1205,6 @@ class InterfacesDict(Dotkeys):
                                  'ifr',
                                  'mode',
                                  'bond_mode',
-                                 'vrf_table',
                                  'address'):
                         if attr in kwarg:
                             device[attr] = kwarg.pop(attr)
@@ -1340,7 +1371,8 @@ class AddressesDict(dict):
                'address': msg.get_attr('IFA_ADDRESS'),
                'flags': msg.get_attr('IFA_FLAGS') or msg.get('flags'),
                'prefixlen': msg['prefixlen'],
-               'family': msg['family']}
+               'family': msg['family'],
+               'cacheinfo': msg.get_attr('IFA_CACHEINFO')}
         try:
             self[msg['index']].add(key=(addr,
                                         raw['prefixlen']),
