@@ -441,11 +441,11 @@ of 'prefix/mask' string and the route priority (if any)::
 
     In [1]: ipdb.routes.tables[254].idx.keys()
     Out[1]:
-    [RouteKey(dst='default', scope=0, table=254, family=2, ...),
-     RouteKey(dst='172.17.0.0/16', scope=253, table=254, ...),
-     RouteKey(dst='172.16.254.0/24', scope=253, table=254, ...),
-     RouteKey(dst='192.168.122.0/24', scope=253, table=254, ...),
-     RouteKey(dst='fe80::/64', scope=0, table=254, family=10, ...)]
+    [RouteKey(dst='default', table=254, family=2, ...),
+     RouteKey(dst='172.17.0.0/16', table=254, ...),
+     RouteKey(dst='172.16.254.0/24', table=254, ...),
+     RouteKey(dst='192.168.122.0/24', table=254, ...),
+     RouteKey(dst='fe80::/64', table=254, family=10, ...)]
 
 But a routing table in IPDB allows several variants of the
 route spec. The simplest case is to retrieve a route by
@@ -703,6 +703,11 @@ import atexit
 import logging
 import traceback
 import threading
+import weakref
+try:
+    import queue
+except ImportError:
+    import Queue as queue  # The module is called 'Queue' in Python2
 
 from functools import partial
 from pprint import pprint
@@ -710,12 +715,13 @@ from pyroute2 import config
 from pyroute2.common import uuid32
 from pyroute2.common import basestring
 from pyroute2.iproute import IPRoute
-from pyroute2.netlink.rtnl import RTM_GETLINK, RTNL_GROUPS
+from pyroute2.netlink.rtnl import RTM_GETLINK, RTMGRP_DEFAULTS
 from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
 from pyroute2.ipdb import rules
 from pyroute2.ipdb import routes
 from pyroute2.ipdb import interfaces
 from pyroute2.ipdb.routes import BaseRoute
+from pyroute2.ipdb.exceptions import ShutdownException
 from pyroute2.ipdb.transactional import SYNC_TIMEOUT
 from pyroute2.ipdb.linkedset import IPaddrSet
 from pyroute2.ipdb.linkedset import SortedIPaddrSet
@@ -754,6 +760,49 @@ class Watchdog(object):
         self.ipdb.unregister_callback(self.uuid)
 
 
+class _evq_context(object):
+    '''
+    Context manager class for the event queue used by the event loop
+    '''
+    def __init__(self, ipdb, qsize, block, timeout):
+        self._ipdb = ipdb
+        self._qsize = qsize
+        self._block = block
+        self._timeout = timeout
+
+    def __enter__(self):
+        # Context manager protocol
+        self._ipdb._evq_lock.acquire()
+        self._ipdb._evq = queue.Queue(maxsize=self._qsize)
+        self._ipdb._evq_drop = 0
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        # Context manager protocol
+        self._ipdb._evq = None
+        self._ipdb._evq_drop = 0
+        self._ipdb._evq_lock.release()
+
+    def __iter__(self):
+        # Iterator protocol
+        if not self._ipdb._evq:
+            raise RuntimeError('eventqueue must be used '
+                               'as a context manager')
+        return self
+
+    def next(self):
+        # Iterator protocol -- Python 2.x compatibility
+        return self.__next__()
+
+    def __next__(self):
+        # Iterator protocol -- Python 3.x
+        msg = self._ipdb._evq.get(self._block, self._timeout)
+        self._ipdb._evq.task_done()
+        if isinstance(msg, Exception):
+            raise msg
+        return msg
+
+
 class IPDB(object):
     '''
     The class that maintains information about network setup
@@ -763,7 +812,8 @@ class IPDB(object):
 
     def __init__(self, nl=None, mode='implicit',
                  restart_on_error=None, nl_async=None,
-                 nl_bind_groups=RTNL_GROUPS,
+                 sndbuf=1048576, rcvbuf=1048576,
+                 nl_bind_groups=RTMGRP_DEFAULTS,
                  ignore_rtables=None, callbacks=None,
                  sort_addresses=False, plugins=None):
         plugins = plugins or ['interfaces', 'routes', 'rules']
@@ -771,6 +821,7 @@ class IPDB(object):
                 'routes': routes,
                 'rules': rules}
         self.mode = mode
+        self.txdrop = False
         self._stdout = sys.stdout
         self._ipaddr_set = SortedIPaddrSet if sort_addresses else IPaddrSet
         self._event_map = {}
@@ -782,6 +833,8 @@ class IPDB(object):
         self._nl_async = config.ipdb_nl_async if nl_async is None else True
         self.mnl = None
         self.nl = nl
+        self._sndbuf = sndbuf
+        self._rcvbuf = rcvbuf
         self.nl_bind_groups = nl_bind_groups
         self._plugins = [pmap[x] for x in plugins if x in pmap]
         if isinstance(ignore_rtables, int):
@@ -794,7 +847,15 @@ class IPDB(object):
         # see also 'register_callback'
         self._post_callbacks = {}
         self._pre_callbacks = {}
-        self._cb_threads = {}
+
+        # local event queues
+        # - callbacks event queue
+        self._cbq = queue.Queue(maxsize=8192)
+        self._cbq_drop = 0
+        # - users event queue
+        self._evq = None
+        self._evq_lock = threading.Lock()
+        self._evq_drop = 0
 
         # locks and events
         self.exclusive = threading.RLock()
@@ -828,11 +889,13 @@ class IPDB(object):
         # init the dir() cache
         self.__dir_cache__ = [i for i in self.__class__.__dict__.keys()
                               if i[0] != '_']
-        self.__dir_cache__.remove('serve_forever')
         self.__dir_cache__.extend(list(self._deferred.keys()))
 
-        #
-        atexit.register(self.release)
+        def cleanup(ref):
+            ipdb_obj = ref()
+            if ipdb_obj is not None:
+                ipdb_obj.release()
+        atexit.register(cleanup, weakref.ref(self))
 
     def __dir__(self):
         return self.__dir_cache__
@@ -843,46 +906,87 @@ class IPDB(object):
     def __exit__(self, exc_type, exc_value, traceback):
         self.release()
 
+    def _flush_db(self):
+
+        def flush(idx):
+            for key in tuple(idx.keys()):
+                try:
+                    del idx[key]
+                except KeyError:
+                    pass
+        idx_list = []
+        if 'interfaces' in self._loaded:
+            for (key, dev) in self.by_name.items():
+                try:
+                    # FIXME
+                    self.interfaces._detach(key,
+                                            dev['index'],
+                                            dev.nlmsg)
+                except KeyError:
+                    pass
+            idx_list.append(self.ipaddr)
+            idx_list.append(self.neighbours)
+        if 'routes' in self._loaded:
+            idx_list.extend([self.routes.tables[x] for x
+                             in self.routes.tables.keys()])
+        if 'rules' in self._loaded:
+            idx_list.append(self.rules)
+        for idx in idx_list:
+            flush(idx)
+
     def initdb(self):
-        # common event map, empty by default, so all the
-        # events aer just ignored
-        self.release(complete=False)
-        self._stop = False
-        # explicitly cleanup object references
-        for event in tuple(self._event_map):
-            del self._event_map[event]
 
-        # if the command socket is not provided, create it
-        if self._nl_own:
-            self.nl = IPRoute()
-        # setup monitoring socket
-        self.mnl = self.nl.clone()
-        try:
-            self.mnl.bind(groups=self.nl_bind_groups, async=self._nl_async)
-        except:
-            self.mnl.close()
-            if self._nl_own is None:
-                self.nl.close()
-            raise
+        # flush all the DB objects
+        with self.exclusive:
 
-        # explicitly cleanup references
-        for key in tuple(self._deferred):
-            del self._deferred[key]
+            # explicitly cleanup object references
+            for event in tuple(self._event_map):
+                del self._event_map[event]
 
-        for module in self._plugins:
-            if (module.groups & self.nl_bind_groups) != module.groups:
-                continue
-            for plugin in module.spec:
-                self._deferred[plugin['name']] = module.spec
-                if plugin['name'] in self._loaded:
-                    delattr(self, plugin['name'])
-                    self._loaded.remove(plugin['name'])
+            self._flush_db()
 
-        # start the monitoring thread
-        self._mthread = threading.Thread(name="IPDB event loop",
-                                         target=self.serve_forever)
-        self._mthread.setDaemon(True)
-        self._mthread.start()
+            # if the command socket is not provided, create it
+            if self._nl_own:
+                if self.nl is not None:
+                    self.nl.close()
+                self.nl = IPRoute(sndbuf=self._sndbuf, rcvbuf=self._rcvbuf)
+            # setup monitoring socket
+            if self.mnl is not None:
+                self._flush_mnl()
+                self.mnl.close()
+            self.mnl = self.nl.clone()
+            try:
+                self.mnl.bind(groups=self.nl_bind_groups,
+                              async_cache=self._nl_async)
+            except:
+                self.mnl.close()
+                if self._nl_own is None:
+                    self.nl.close()
+                raise
+
+            # explicitly cleanup references
+            for key in tuple(self._deferred):
+                del self._deferred[key]
+
+            for module in self._plugins:
+                if (module.groups & self.nl_bind_groups) != module.groups:
+                    continue
+                for plugin in module.spec:
+                    self._deferred[plugin['name']] = module.spec
+                    if plugin['name'] in self._loaded:
+                        delattr(self, plugin['name'])
+                        self._loaded.remove(plugin['name'])
+
+            # start service threads
+            for tspec in (('_mthread', '_serve_main', 'IPDB main event loop'),
+                          ('_cthread', '_serve_cb', 'IPDB cb event loop')):
+                tg = getattr(self, tspec[0], None)
+                if not getattr(tg, 'is_alive', lambda: False)():
+                    tx = threading.Thread(name=tspec[2],
+                                          target=getattr(self, tspec[1]))
+                    setattr(self, tspec[0], tx)
+                    tx.setDaemon(True)
+                    tx.start()
 
     def __getattribute__(self, name):
         deferred = super(IPDB, self).__getattribute__('_deferred')
@@ -988,13 +1092,36 @@ class IPDB(object):
             raise KeyError('Unknown callback mode')
         safe = cbchain[cuid]
         with safe.lock:
-            cbchain.pop(cuid)
-        for t in tuple(self._cb_threads.get(cuid, ())):
-            t.join(3)
-        ret = self._cb_threads.get(cuid, ())
+            ret = cbchain.pop(cuid)
         return ret
 
-    def release(self, complete=True):
+    def eventqueue(self, qsize=8192, block=True, timeout=None):
+        '''
+        Initializes event queue and returns event queue context manager.
+        Once the context manager is initialized, events start to be collected,
+        so it is possible to read initial state from the system witout losing
+        last moment changes, and once that is done, start processing events.
+
+        Example:
+
+            ipdb = IPDB()
+            with ipdb.eventqueue() as evq:
+                my_state = ipdb.<needed_attribute>...
+                for msg in evq:
+                    update_state_by_msg(my_state, msg)
+        '''
+        return _evq_context(self, qsize, block, timeout)
+
+    def eventloop(self, qsize=8192, block=True, timeout=None):
+        """
+        Event generator for simple cases when there is no need for initial
+        state setup. Initialize event queue and yield events as they happen.
+        """
+        with self.eventqueue(qsize=qsize, block=block, timeout=timeout) as evq:
+            for msg in evq:
+                yield msg
+
+    def release(self):
         '''
         Shutdown IPDB instance and sync the state. Since
         IPDB is asyncronous, some operations continue in the
@@ -1008,67 +1135,38 @@ class IPDB(object):
         '''
         with self._shutdown_lock:
             if self._stop:
+                log.warning("shutdown in progress")
                 return
             self._stop = True
-            if self.mnl is not None:
-                # terminate the main loop
-                for t in range(3):
-                    try:
-                        msg = ifinfmsg()
-                        msg['index'] = 1
-                        msg.reset()
-                        self.mnl.put(msg, RTM_GETLINK)
-                    except Exception as e:
-                        log.warning("shutdown error: %s", e)
-                        # Just give up.
-                        # We can not handle this case
+            self._cbq.put(ShutdownException("shutdown"))
 
             if self._mthread is not None:
+                self._flush_mnl()
                 self._mthread.join()
 
             if self.mnl is not None:
                 self.mnl.close()
                 self.mnl = None
-                if complete or self._nl_own:
-                    self.nl.close()
-                    self.nl = None
 
-        with self.exclusive:
-                # collect all the callbacks
-                for cuid in tuple(self._cb_threads):
-                    for t in tuple(self._cb_threads[cuid]):
-                        t.join()
+            if self._nl_own:
+                self.nl.close()
+                self.nl = None
 
-                # flush all the objects
-                def flush(idx):
-                    for key in tuple(idx.keys()):
-                        try:
-                            del idx[key]
-                        except KeyError:
-                            pass
-                idx_list = []
+            self._flush_db()
 
-                if 'interfaces' in self._loaded:
-                    for (key, dev) in self.by_name.items():
-                        try:
-                            # FIXME
-                            self.interfaces._detach(key,
-                                                    dev['index'],
-                                                    dev.nlmsg)
-                        except KeyError:
-                            pass
-                    idx_list.append(self.ipaddr)
-                    idx_list.append(self.neighbours)
-
-                if 'routes' in self._loaded:
-                    idx_list.extend([self.routes.tables[x] for x
-                                     in self.routes.tables.keys()])
-
-                if 'rules' in self._loaded:
-                    idx_list.append(self.rules)
-
-                for idx in idx_list:
-                    flush(idx)
+    def _flush_mnl(self):
+        if self.mnl is not None:
+            # terminate the main loop
+            for t in range(3):
+                try:
+                    msg = ifinfmsg()
+                    msg['index'] = 1
+                    msg.reset()
+                    self.mnl.put(msg, RTM_GETLINK)
+                except Exception as e:
+                    log.error("shutdown error: %s", e)
+                    # Just give up.
+                    # We can not handle this case
 
     def create(self, kind, ifname, reuse=False, **kwarg):
         return self.interfaces.add(kind, ifname, reuse, **kwarg)
@@ -1205,7 +1303,8 @@ class IPDB(object):
             # 8<------------------------------
             # intefaces
             kind = target.get('kind', None)
-            if kind in ('vlan', 'vxlan', 'gre', 'tuntap', 'vti', 'vrf'):
+            if kind in ('vlan', 'vxlan', 'gre', 'tuntap', 'vti', 'vti6',
+                        'vrf'):
                 tx_prio1.append((target, tx))
             elif kind in ('bridge', 'bond'):
                 tx_prio2.append((target, tx))
@@ -1268,16 +1367,32 @@ class IPDB(object):
 
         return self
 
-    def watchdog(self, action='RTM_NEWLINK', **kwarg):
-        return Watchdog(self, action, kwarg)
+    def watchdog(self, wdops='RTM_NEWLINK', **kwarg):
+        return Watchdog(self, wdops, kwarg)
 
-    def serve_forever(self):
+    def _serve_cb(self):
+        ###
+        # Callbacks thread working on a dedicated event queue.
+        ###
+
+        while not self._stop:
+            msg = self._cbq.get()
+            self._cbq.task_done()
+            if isinstance(msg, ShutdownException):
+                return
+            elif isinstance(msg, Exception):
+                raise msg
+            for cb in tuple(self._post_callbacks.values()):
+                try:
+                    cb(self, msg, msg['event'])
+                except:
+                    pass
+
+    def _serve_main(self):
         ###
         # Main monitoring cycle. It gets messages from the
         # default iproute queue and updates objects in the
         # database.
-        #
-        # Should not be called manually.
         ###
 
         while not self._stop:
@@ -1292,10 +1407,14 @@ class IPDB(object):
                 # anymore
                 if self._stop:
                     break
-            except:
-                log.error('Restarting IPDB instance after '
-                          'error:\n%s', traceback.format_exc())
+            except Exception as e:
+                with self.exclusive:
+                    if self._evq:
+                        self._evq.put(e)
+                        return
                 if self.restart_on_error:
+                    log.error('Restarting IPDB instance after '
+                              'error:\n%s', traceback.format_exc())
                     try:
                         self.initdb()
                     except:
@@ -1304,6 +1423,7 @@ class IPDB(object):
                         return
                     continue
                 else:
+                    log.error('Emergency shutdown, cleanup manually')
                     raise RuntimeError('Emergency shutdown')
 
             for msg in messages:
@@ -1321,25 +1441,35 @@ class IPDB(object):
                         for func in self._event_map[event]:
                             func(msg)
 
-                # run post-callbacks
-                # NOTE: post-callbacks are asynchronous
-                for (cuid, cb) in tuple(self._post_callbacks.items()):
-                    t = threading.Thread(name="IPDB callback %s" % (id(cb)),
-                                         target=cb,
-                                         args=(self, msg, msg['event']))
-                    t.start()
-                    if cuid not in self._cb_threads:
-                        self._cb_threads[cuid] = set()
-                    self._cb_threads[cuid].add(t)
+                    # Post-callbacks
+                    try:
+                        self._cbq.put_nowait(msg)
+                        if self._cbq_drop:
+                            log.warning('dropped %d events',
+                                        self._cbq_drop)
+                            self._cbq_drop = 0
+                    except queue.Full:
+                        self._cbq_drop += 1
+                    except Exception:
+                        log.error('Emergency shutdown, cleanup manually')
+                        raise RuntimeError('Emergency shutdown')
 
-                # occasionally join cb threads
-                for cuid in tuple(self._cb_threads):
-                    for t in tuple(self._cb_threads.get(cuid, ())):
-                        t.join(0)
-                        if not t.is_alive():
-                            try:
-                                self._cb_threads[cuid].remove(t)
-                            except KeyError:
-                                pass
-                            if len(self._cb_threads.get(cuid, ())) == 0:
-                                del self._cb_threads[cuid]
+                    #
+                    # Why not to put these two pieces of the code
+                    # it in a routine?
+                    #
+                    # TODO: run performance tests with routines
+
+                    # Users event queue
+                    if self._evq:
+                        try:
+                            self._evq.put_nowait(msg)
+                            if self._evq_drop:
+                                log.warning("dropped %d events",
+                                            self._evq_drop)
+                                self._evq_drop = 0
+                        except queue.Full:
+                            self._evq_drop += 1
+                        except Exception as e:
+                            log.error('Emergency shutdown, cleanup manually')
+                            raise RuntimeError('Emergency shutdown')
