@@ -2,6 +2,7 @@ import time
 import types
 import struct
 import logging
+import traceback
 import threading
 from collections import namedtuple
 from socket import AF_UNSPEC
@@ -14,6 +15,8 @@ from pyroute2.common import basestring
 from pyroute2.netlink import rtnl
 from pyroute2.netlink import nlmsg
 from pyroute2.netlink import nlmsg_base
+from pyroute2.netlink import NLM_F_MULTI
+from pyroute2.netlink import NLM_F_CREATE
 from pyroute2.netlink.rtnl import rt_type
 from pyroute2.netlink.rtnl import rt_proto
 from pyroute2.netlink.rtnl import encap_type
@@ -27,9 +30,9 @@ from pyroute2.ipdb.transactional import SYNC_TIMEOUT
 from pyroute2.ipdb.linkedset import LinkedSet
 
 log = logging.getLogger(__name__)
-groups = rtnl.RTNLGRP_IPV4_ROUTE |\
-    rtnl.RTNLGRP_IPV6_ROUTE |\
-    rtnl.RTNLGRP_MPLS_ROUTE
+groups = rtnl.RTMGRP_IPV4_ROUTE |\
+    rtnl.RTMGRP_IPV6_ROUTE |\
+    rtnl.RTMGRP_MPLS_ROUTE
 IP6_RT_PRIO_USER = 1024
 
 
@@ -84,9 +87,9 @@ class NextHopSet(LinkedSet):
 
     def add(self, prime, raw=None, cascade=False):
         key = self.__make_nh(prime)
-        r = key._required
-        l = key._fields
-        skey = key[:r] + (None, ) * (len(l) - r)
+        req = key._required
+        fields = key._fields
+        skey = key[:req] + (None, ) * (len(fields) - req)
         if skey in self.raw:
             del self.raw[skey]
         return super(NextHopSet, self).add(key, raw=prime)
@@ -96,11 +99,11 @@ class NextHopSet(LinkedSet):
         try:
             super(NextHopSet, self).remove(key)
         except KeyError as e:
-            r = key._required
-            l = key._fields
-            skey = key[:r] + (None, ) * (len(l) - r)
+            req = key._required
+            fields = key._fields
+            skey = key[:req] + (None, ) * (len(fields) - req)
             for rkey in tuple(self.raw.keys()):
-                if skey == rkey[:r] + (None, ) * (len(l) - r):
+                if skey == rkey[:req] + (None, ) * (len(fields) - req):
                     break
             else:
                 raise e
@@ -126,21 +129,26 @@ class WatchdogKey(dict):
                                          'dst_len',
                                          'src',
                                          'src_len',
-                                         'oif',
-                                         'iif',
+                                         'tos',
+                                         'priority',
                                          'gateway',
                                          'table') and x[1]])
 
 
 # Universal route key
+# Holds the fields that the kernel uses to uniquely identify routes.
+# IPv4 allows redundant routes with different 'tos' but IPv6 does not,
+# so 'tos' is used for IPv4 but not IPv6.
+# For reference, see fib_table_insert() in
+# https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/net/ipv4/fib_trie.c#n1147
+# and fib6_add_rt2node() in
+# https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/net/ipv6/ip6_fib.c#n765
 RouteKey = namedtuple('RouteKey',
                       ('dst',
                        'table',
                        'family',
                        'priority',
-                       'tos',
-                       'oif'))
-RouteKey._required = 5  # number of required fields (should go first)
+                       'tos'))
 
 # IP multipath NH key
 IPNHKey = namedtuple('IPNHKey',
@@ -213,23 +221,7 @@ class BaseRoute(Transactional):
             # nexthop as the first in the multipath
             if not self['multipath']:
                 first = {}
-                for key in [x for x in self._fields
-                            if x not in ('family',
-                                         'dst_len',
-                                         'src_len',
-                                         'tos',
-                                         'table',
-                                         'proto',
-                                         'scope',
-                                         'type',
-                                         'flags',
-                                         'dst',
-                                         'src',
-                                         'ipdb_scope',
-                                         'metrics',
-                                         'encap',
-                                         'via',
-                                         'multipath')]:
+                for key in ('oif', 'gateway', 'newdst'):
                     if self[key]:
                         first[key] = self[key]
                 if first:
@@ -241,7 +233,7 @@ class BaseRoute(Transactional):
                             self[key] = None
                     self['multipath'].add(first)
                     # cleanup key fields
-                    for key in ('oif', 'iif', 'gateway', 'newdst'):
+                    for key in ('oif', 'gateway', 'newdst'):
                         self[key] = None
             # add the prime as NH
             if self['family'] == AF_MPLS:
@@ -269,11 +261,55 @@ class BaseRoute(Transactional):
             for (key, value) in msg.items():
                 self[key] = value
 
+            # IPv6 multipath via several devices (not networks) is a very
+            # special case, since we get only the first hop notification. Ask
+            # the kernel guys why. I've got no idea.
+            #
+            # So load all the rest
+            flags = msg.get('header', {}).get('flags', 0)
+            family = msg.get('family', 0)
+            clean_mp = True
+            #
+            # It MAY be a multipath hop
+            #
+            if family == AF_INET6 and not msg.get_attr('RTA_MULTIPATH'):
+                #
+                # It is a notification about the route created
+                #
+                if flags == NLM_F_CREATE:
+                    #
+                    # This routine can significantly slow down the IPDB
+                    # instance, but I see no way around. Some are born
+                    # to endless night.
+                    #
+                    clean_mp = False
+                    msgs = self.nl.route('show',
+                                         table=self['table'],
+                                         dst=self['dst'],
+                                         family=self['family'])
+                    for nhmsg in msgs:
+                        nh = type(self)(ipdb=self.ipdb, parent=self)
+                        nh.load_netlink(nhmsg)
+                        with nh._direct_state:
+                            del nh['dst']
+                            del nh['ipdb_scope']
+                            del nh['ipdb_priority']
+                            del nh['multipath']
+                            del nh['metrics']
+                        self.add_nh(nh)
+                #
+                # it IS a multipath hop loaded during IPDB init
+                #
+                elif flags == NLM_F_MULTI and self.get('dst'):
+                    nh = type(self)(ipdb=self.ipdb, parent=self)
+                    nh.load_netlink(msg)
+                    return
+
             # cleanup multipath NH
-            for nh in self['multipath']:
-                self.del_nh(nh)
+            if clean_mp:
+                for nh in self['multipath']:
+                    self.del_nh(nh)
 
-            # merge NLA
             for cell in msg['attrs']:
                 #
                 # Parse on demand
@@ -304,9 +340,14 @@ class BaseRoute(Transactional):
                 elif norm == 'encap':
                     with self['encap']._direct_state:
                         ret = []
-                        for l in value.get_attr('MPLS_IPTUNNEL_DST'):
-                            ret.append(str(l['label']))
-                        self['encap']['labels'] = '/'.join(ret)
+                        # FIXME: should support encap_types other than MPLS
+                        try:
+                            for l in value.get_attr('MPLS_IPTUNNEL_DST'):
+                                ret.append(str(l['label']))
+                            if ret:
+                                self['encap']['labels'] = '/'.join(ret)
+                        except AttributeError:
+                            pass
                 elif norm == 'via':
                     with self['via']._direct_state:
                         self['via'] = value
@@ -372,17 +413,21 @@ class BaseRoute(Transactional):
             return self
 
         error = None
-        drop = True
+        drop = self.ipdb.txdrop
         devop = 'set'
         cleanup = []
+        # FIXME -- make a debug object
+        debug = {'traceback': None,
+                 'next_stage': None}
+        notx = True
+
+        if tid or transaction:
+            notx = False
 
         if tid:
             transaction = self.global_tx[tid]
         else:
-            if transaction:
-                drop = False
-            else:
-                transaction = self.current_tx
+            transaction = transaction or self.current_tx
 
         # ignore global rollbacks on invalid routes
         if self['ipdb_scope'] == 'create' and commit_phase > 1:
@@ -416,7 +461,7 @@ class BaseRoute(Transactional):
                     if mplen == 1:
                         # set up local targets
                         for nh in transaction['multipath']:
-                            for key in ('gateway', 'oif', 'newdst'):
+                            for key in ('oif', 'gateway', 'newdst'):
                                 if nh.get(key, None):
                                     self.set_target(key, nh[key])
                                     wlist.append(key)
@@ -433,9 +478,6 @@ class BaseRoute(Transactional):
                 old_key = self.make_key(self)
                 new_key = self.make_key(transaction)
                 if old_key != new_key:
-                    # delete old record
-                    if devop == 'set':
-                        self.nl.route('del', **dict(old_key._asdict()))
                     # assume we can not move routes between tables (yet ;)
                     if self['family'] == AF_MPLS:
                         route_index = self.ipdb.routes.tables['mpls'].idx
@@ -454,6 +496,16 @@ class BaseRoute(Transactional):
                     if old_key in route_index:
                         del route_index[old_key]
                 self.nl.route(devop, **transaction)
+                # delete old record, if required
+                if (old_key != new_key) and (devop == 'set'):
+                    req = dict(old_key._asdict())
+                    # update the request with the scope.
+                    #
+                    # though the scope isn't a part of the
+                    # key, it is required for the correct
+                    # removal -- only if it is set
+                    req['scope'] = self.get('scope', 0)
+                    self.nl.route('del', **req)
                 transaction.wait_all_targets()
                 for key in ('metrics', 'via'):
                     if transaction[key] and transaction[key]._targets:
@@ -482,27 +534,31 @@ class BaseRoute(Transactional):
                     with self._direct_state:
                         self['ipdb_scope'] = 'shadow'
 
-        except Exception as e:
-            if commit_phase == 1:
-                ret = self.commit(transaction=snapshot,
-                                  commit_phase=2,
-                                  commit_mask=commit_mask)
-                if isinstance(ret, Exception):
-                    error = ret
-                else:
-                    error = e
-            else:
-                if drop:
-                    self.drop(transaction.uid)
-                x = RuntimeError()
-                x.cause = e
-                raise x
+            # success, so it's safe to drop the transaction
+            drop = True
 
-        if drop and commit_phase == 1:
+        except Exception as e:
+
+            error = e
+            # prepare postmortem
+            debug['traceback'] = traceback.format_exc()
+            debug['error_stack'] = []
+            debug['next_stage'] = None
+
+            if commit_phase == 1:
+                try:
+                    self.commit(transaction=snapshot,
+                                commit_phase=2,
+                                commit_mask=commit_mask)
+                except Exception as i_e:
+                    debug['next_stage'] = i_e
+                    error = RuntimeError()
+
+        if drop and notx:
             self.drop(transaction.uid)
 
         if error is not None:
-            error.transaction = transaction
+            error.debug = debug
             raise error
 
         self.ipdb.routes.gc()
@@ -606,7 +662,7 @@ class Route(BaseRoute):
                     else:
                         v = 'default'
                 elif field == 'tos' and msg.get('family') != AF_INET:
-                    # ignore tos field for non-IPv4 routes,
+                    # ignore tos field for non-IPv6 routes,
                     # as it used as a key only there
                     v = None
                 elif v is None:
@@ -625,7 +681,7 @@ class Route(BaseRoute):
                     else:
                         v = ip
                 elif field == 'tos' and msg.get('family') != AF_INET:
-                    # ignore tos field for non-IPv4 routes,
+                    # ignore tos field for non-IPv6 routes,
                     # as it used as a key only there
                     v = None
                 values.append(v)
@@ -685,6 +741,10 @@ class Route(BaseRoute):
             ret = rt_type.get(value, value)
         elif key == 'proto' and not isinstance(value, int):
             ret = rt_proto.get(value, value)
+        elif key == 'dst' and \
+                isinstance(value, basestring) and \
+                value in ('0.0.0.0/0', '::/0'):
+            ret = 'default'
         Transactional.__setitem__(self, key, ret)
 
     def __getitem__(self, key):
@@ -823,7 +883,7 @@ class RoutingTable(object):
     def filter(self, target, oneshot=False):
         #
         if isinstance(target, types.FunctionType):
-            return filter(target, [x for x in self.idx.values()])
+            return filter(target, [x for x in tuple(self.idx.values())])
 
         if isinstance(target, basestring):
             target = {'dst': target}
@@ -832,8 +892,8 @@ class RoutingTable(object):
             raise TypeError('target type not supported: %s' % type(target))
 
         ret = []
-        for record in self.idx.values():
-            for key, value in target.items():
+        for record in tuple(self.idx.values()):
+            for key, value in tuple(target.items()):
                 if (key not in record['route']) or \
                         (value != record['route'][key]):
                     break
@@ -853,27 +913,8 @@ class RoutingTable(object):
 
         # match the route by key
         if isinstance(target, (tuple, list)):
-            try:
-                # full match
-                return self.idx[RouteKey(*target)]
-            except KeyError:
-                # w/o iif and oif
-                # when a route is just created, there can be no oif and
-                # iif specified, if they weren't provided explicitly,
-                # and in that case there will be the key w/o oif and
-                # iif
-                r = RouteKey._required
-                l = RouteKey._fields
-                try:
-                    return self.idx[RouteKey(*(target[:r] +
-                                               (None, ) * (len(l) - r)))]
-                except KeyError:
-                    # no luck with keys, fallback to simple search
-                    for key in self.idx:
-                        if key[:r] == target[:r]:
-                            return self.idx[key]
-                    else:
-                        raise
+            # full match
+            return self.idx[RouteKey(*target)]
 
         if isinstance(target, nlmsg):
             return self.idx[Route.make_key(target)]
@@ -1003,6 +1044,8 @@ class RoutingTableSet(object):
         Create a route from a dictionary
         '''
         spec = dict(spec or kwarg)
+        gateway = spec.get('gateway') or ''
+        dst = spec.get('dst') or ''
         if 'tos' not in spec:
             spec['tos'] = 0
         if 'scope' not in spec:
@@ -1010,12 +1053,11 @@ class RoutingTableSet(object):
         if 'table' not in spec:
             spec['table'] = 254
         if 'family' not in spec:
-            if (spec.get('dst', '').find(':') > -1) or \
-                    (spec.get('gateway', '').find(':') > -1):
+            if (dst.find(':') > -1) or (gateway.find(':') > -1):
                 spec['family'] = AF_INET6
             else:
                 spec['family'] = AF_INET
-        if 'dst' not in spec:
+        if not dst:
             raise ValueError('dst not specified')
         if 'priority' not in spec:
             if spec['family'] == AF_INET6:
@@ -1092,12 +1134,11 @@ class RoutingTableSet(object):
                 self.tables[table] = MPLSTable(self.ipdb)
             else:
                 self.tables[table] = RoutingTable(self.ipdb)
-        key = self.tables[table].load(msg)
-        return self.tables[table][key]
+        self.tables[table].load(msg)
 
     def gc_mark_addr(self, msg):
         ##
-        # Find invalid route records after addr delete
+        # Find invalid IPv4 route records after addr delete
         #
         # Example::
         #   $ sudo ip link add test0 type dummy
@@ -1109,9 +1150,13 @@ class RoutingTableSet(object):
         #
         # The route {'dst': '10.1.2.0/24', 'gateway': '172.18.0.1'}
         # will stay in the routing table being removed from the system.
-        # That's because the kernel doesn't send route updates in that
-        # case, so we have to calculate the update here -- or load all
-        # the routes from scratch. The latter may be far too expensive.
+        # That's because the kernel doesn't send IPv4 route updates in
+        # that case, so we have to calculate the update here -- or load
+        # all the routes from scratch. The latter may be far too
+        # expensive.
+        #
+        # See http://www.spinics.net/lists/netdev/msg254186.html for
+        # background on this kernel behavior.
 
         # Simply ignore secondary addresses, as they don't matter
         if msg['flags'] & IFA_F_SECONDARY:
@@ -1144,7 +1189,8 @@ class RoutingTableSet(object):
                             record['route']._gctime = time.time()
 
         elif family == AF_INET6:
-            # TODO: add IPv6 support
+            # Unlike IPv4, IPv6 route updates are sent after addr
+            # delete, so no need to delete them here.
             pass
         else:
             # ignore not (IPv4 or IPv6)
@@ -1181,7 +1227,7 @@ class RoutingTableSet(object):
     def filter(self, target):
         # FIXME: turn into generator!
         ret = []
-        for table in self.tables.values():
+        for table in tuple(self.tables.values()):
             if table is not None:
                 ret.extend(table.filter(target))
         return ret
