@@ -195,7 +195,7 @@ You can also make your own NLA descriptions::
                    ...)
 
         class cacheinfo(nla):
-            fields = (('ifa_prefered', 'I'),
+            fields = (('ifa_preferred', 'I'),
                       ('ifa_valid', 'I'),
                       ('cstamp', 'I'),
                       ('tstamp', 'I'))
@@ -591,6 +591,8 @@ class nlmsg_base(dict):
     cell_header = None
     align = 4
     nla_map = {}                 # NLA mapping
+    sql_constraints = {}
+    sql_extra_fields = tuple()
     nla_flags = 0        # NLA flags
     value_map = {}
     is_nla = False
@@ -662,6 +664,32 @@ class nlmsg_base(dict):
         if self.header:
             self['header'] = {}
 
+    @classmethod
+    def sql_schema(cls):
+        ret = []
+        for field in cls.fields:
+            if field[0][0] != '_':
+                ret.append((field[0],
+                            ' '.join(('INTEGER',
+                                      cls.sql_constraints.get(field[0], '')))))
+        for nla in cls.nla_map:
+            if isinstance(nla[0], basestring):
+                nla_name = nla[0]
+                nla_type = nla[1]
+            else:
+                nla_name = nla[1]
+                nla_type = nla[2]
+            nla_type = getattr(cls, nla_type, None)
+            sql_type = getattr(nla_type, 'sql_type', None)
+            if sql_type:
+                sql_type = ' '.join((sql_type,
+                                     cls.sql_constraints.get(nla_name, '')))
+                ret.append((nla_name, sql_type))
+
+        for field in cls.sql_extra_fields:
+            ret.append(field)
+        return ret
+
     @property
     def buf(self):
         logging.error('nlmsg.buf is deprecated:\n%s',
@@ -903,7 +931,7 @@ class nlmsg_base(dict):
             # If the key exists, the statement after the first `or` is not
             # executed.
             if self.is_nla:
-                key = tuple(self.data[offset:offset+4])
+                key = tuple(self.data[offset:offset + 4])
                 self['header'] = cache_hdr.get(key, None) or \
                     (cache_hdr
                      .__setitem__(key,
@@ -975,6 +1003,7 @@ class nlmsg_base(dict):
                                   offset=offset,
                                   parent=self)
                 cell._nla_array = False
+                cell['header']['type'] = 1
                 if cell.cell_header is not None:
                     cell.header = cell.cell_header
                 cell.setvalue(value)
@@ -1170,9 +1199,10 @@ class nlmsg_base(dict):
 
     @staticmethod
     def _ft_decode_zstring(self, offset):
-        self['value'], = struct.unpack_from('%is' % (self.length - 5),
-                                            self.data,
-                                            offset)
+        value, = struct.unpack_from('%is' % (self.length - 4),
+                                    self.data,
+                                    offset)
+        self['value'] = value.strip(b'\0')
 
     @staticmethod
     def _ft_decode_string(self, offset):
@@ -1416,7 +1446,7 @@ class nla_slot(object):
         if self.try_to_decode():
             return cell.getvalue()
         else:
-            return cell.data[cell.offset:cell.offset+cell.length]
+            return cell.data[cell.offset:cell.offset + cell.length]
 
     def get_flags(self):
         if self.try_to_decode():
@@ -1489,60 +1519,91 @@ class nlmsg_atoms(nlmsg_base):
     class uint8(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', 'B')]
 
     class uint16(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', 'H')]
 
     class uint32(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', 'I')]
 
     class uint64(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', 'Q')]
 
+    class int8(nla_base):
+
+        __slots__ = ()
+        sql_type = 'INTEGER'
+
+        fields = [('value', 'b')]
+
+    class int16(nla_base):
+
+        __slots__ = ()
+        sql_type = 'INTEGER'
+
+        fields = [('value', 'h')]
+
     class int32(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', 'i')]
 
+    class int64(nla_base):
+
+        __slots__ = ()
+        sql_type = 'INTEGER'
+
+        fields = [('value', 'q')]
+
     class be8(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', '>B')]
 
     class be16(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', '>H')]
 
     class be32(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', '>I')]
 
     class be64(nla_base):
 
         __slots__ = ()
+        sql_type = 'INTEGER'
 
         fields = [('value', '>Q')]
 
     class ipXaddr(nla_base):
 
         __slots__ = ()
+        sql_type = 'TEXT'
 
         fields = [('value', 's')]
         family = None
@@ -1584,6 +1645,7 @@ class nlmsg_atoms(nlmsg_base):
         '''
 
         __slots__ = ()
+        sql_type = 'TEXT'
 
         fields = [('value', 's')]
 
@@ -1616,6 +1678,7 @@ class nlmsg_atoms(nlmsg_base):
         '''
 
         __slots__ = ()
+        sql_type = 'TEXT'
 
         fields = [('value', 's')]
         family = None
@@ -1663,7 +1726,8 @@ class nlmsg_atoms(nlmsg_base):
             elif family == AF_MPLS:
                 self.value = []
                 for i in range(len(self['value']) // 4):
-                    label = struct.unpack('>I', self['value'][i*4:i*4+4])[0]
+                    label = struct.unpack('>I',
+                                          self['value'][i * 4:i * 4 + 4])[0]
                     record = {'label': (label & 0xFFFFF000) >> 12,
                               'tc': (label & 0x00000E00) >> 9,
                               'bos': (label & 0x00000100) >> 8,
@@ -1684,6 +1748,7 @@ class nlmsg_atoms(nlmsg_base):
         '''
 
         __slots__ = ()
+        sql_type = 'TEXT'
 
         fields = [('value', '=6s')]
 
@@ -1738,7 +1803,7 @@ class nlmsg_atoms(nlmsg_base):
 
         def encode(self):
             fmt = '%s%i%s' % (self.fmt[:-1], len(self.value), self.fmt[-1:])
-            self['value'] = struct.pack(fmt, self.value)
+            self['value'] = struct.pack(fmt, *self.value)
             nla_base.encode(self)
 
         def decode(self):
@@ -1766,6 +1831,7 @@ class nlmsg_atoms(nlmsg_base):
         '''
 
         __slots__ = ()
+        sql_type = 'TEXT'
 
         fields = [('value', 's')]
 
