@@ -4,6 +4,7 @@ NL80211 module
 
 TODO
 '''
+import struct
 from pyroute2.common import map_namespace
 from pyroute2.netlink import genlmsg
 from pyroute2.netlink.generic import GenericNetlinkSocket
@@ -11,11 +12,6 @@ from pyroute2.netlink.nlsocket import Marshal
 from pyroute2.netlink import nla
 from pyroute2.netlink import nla_base
 
-
-# import pdb
-import struct
-from pyroute2.common import hexdump
-
 # nl80211 commands
 
 NL80211_CMD_UNSPEC = 0
@@ -144,6 +140,7 @@ NL80211_BSS_ELEMENTS_SSID = 0
 NL80211_BSS_ELEMENTS_SUPPORTED_RATES = 1
 NL80211_BSS_ELEMENTS_CHANNEL = 3
 NL80211_BSS_ELEMENTS_TIM = 5
+NL80211_BSS_ELEMENTS_RSN = 48
 NL80211_BSS_ELEMENTS_EXTENDED_RATE = 50
 NL80211_BSS_ELEMENTS_VENDOR = 221
 
@@ -180,6 +177,35 @@ NL80211_CHAN_WIDTH_10 = 7       # 10 MHz OFDM channel
                                            globals(),
                                            normalize=True)
 
+# BSS "status"
+NL80211_BSS_STATUS_AUTHENTICATED = 0  # Authenticated with this BS
+NL80211_BSS_STATUS_ASSOCIATED = 1     # Associated with this BSS
+NL80211_BSS_STATUS_IBSS_JOINED = 2    # Joined to this IBSS
+(BSS_STATUS_NAMES, BSS_STATUS_VALUES) = map_namespace('NL80211_BSS_STATUS_',
+                                                      globals(),
+                                                      normalize=True)
+
+NL80211_SCAN_FLAG_LOW_PRIORITY = 1 << 0
+NL80211_SCAN_FLAG_FLUSH = 1 << 1
+NL80211_SCAN_FLAG_AP = 1 << 2
+NL80211_SCAN_FLAG_RANDOM_ADDR = 1 << 3
+NL80211_SCAN_FLAG_FILS_MAX_CHANNEL_TIME = 1 << 4
+NL80211_SCAN_FLAG_ACCEPT_BCAST_PROBE_RESP = 1 << 5
+NL80211_SCAN_FLAG_OCE_PROBE_REQ_HIGH_TX_RATE = 1 << 6
+NL80211_SCAN_FLAG_OCE_PROBE_REQ_DEFERRAL_SUPPRESSION = 1 << 7
+(SCAN_FLAGS_NAMES, SCAN_FLAGS_VALUES) = map_namespace('NL80211_SCAN_FLAG_',
+                                                      globals())
+
+NL80211_STA_FLAG_AUTHORIZED = 1
+NL80211_STA_FLAG_SHORT_PREAMBLE = 2
+NL80211_STA_FLAG_WME = 3
+NL80211_STA_FLAG_MFP = 4
+NL80211_STA_FLAG_AUTHENTICATED = 5
+NL80211_STA_FLAG_TDLS_PEER = 6
+NL80211_STA_FLAG_ASSOCIATED = 7
+(STA_FLAG_NAMES, STA_FLAG_VALUES) = map_namespace('NL80211_STA_FLAG_',
+                                                  globals())
+
 
 class nl80211cmd(genlmsg):
     prefix = 'NL80211_ATTR_'
@@ -204,7 +230,7 @@ class nl80211cmd(genlmsg):
                ('NL80211_ATTR_STA_LISTEN_INTERVAL', 'hex'),
                ('NL80211_ATTR_STA_SUPPORTED_RATES', 'hex'),
                ('NL80211_ATTR_STA_VLAN', 'hex'),
-               ('NL80211_ATTR_STA_INFO', 'hex'),
+               ('NL80211_ATTR_STA_INFO', 'STAInfo'),
                ('NL80211_ATTR_WIPHY_BANDS', 'hex'),
                ('NL80211_ATTR_MNTR_FLAGS', 'hex'),
                ('NL80211_ATTR_MESH_ID', 'hex'),
@@ -228,7 +254,7 @@ class nl80211cmd(genlmsg):
                ('NL80211_ATTR_IE', 'hex'),
                ('NL80211_ATTR_MAX_NUM_SCAN_SSIDS', 'uint8'),
                ('NL80211_ATTR_SCAN_FREQUENCIES', 'hex'),
-               ('NL80211_ATTR_SCAN_SSIDS', 'hex'),
+               ('NL80211_ATTR_SCAN_SSIDS', '*string'),
                ('NL80211_ATTR_GENERATION', 'uint32'),
                ('NL80211_ATTR_BSS', 'bss'),
                ('NL80211_ATTR_REG_INITIATOR', 'hex'),
@@ -236,8 +262,8 @@ class nl80211cmd(genlmsg):
                ('NL80211_ATTR_SUPPORTED_COMMANDS', 'hex'),
                ('NL80211_ATTR_FRAME', 'hex'),
                ('NL80211_ATTR_SSID', 'string'),
-               ('NL80211_ATTR_AUTH_TYPE', 'hex'),
-               ('NL80211_ATTR_REASON_CODE', 'hex'),
+               ('NL80211_ATTR_AUTH_TYPE', 'uint32'),
+               ('NL80211_ATTR_REASON_CODE', 'uint16'),
                ('NL80211_ATTR_KEY_TYPE', 'hex'),
                ('NL80211_ATTR_MAX_SCAN_IE_LEN', 'uint16'),
                ('NL80211_ATTR_CIPHER_SUITES', 'hex'),
@@ -341,7 +367,7 @@ class nl80211cmd(genlmsg):
                ('NL80211_ATTR_CONN_FAILED_REASON', 'hex'),
                ('NL80211_ATTR_SAE_DATA', 'hex'),
                ('NL80211_ATTR_VHT_CAPABILITY', 'hex'),
-               ('NL80211_ATTR_SCAN_FLAGS', 'hex'),
+               ('NL80211_ATTR_SCAN_FLAGS', 'uint32'),
                ('NL80211_ATTR_CHANNEL_WIDTH', 'uint32'),
                ('NL80211_ATTR_CENTER_FREQ1', 'uint32'),
                ('NL80211_ATTR_CENTER_FREQ2', 'uint32'),
@@ -405,52 +431,36 @@ class nl80211cmd(genlmsg):
     class bss(nla):
         class elementsBinary(nla_base):
 
-            def binary_rates(self, rawdata):
-                # pdb.set_trace()
+            def binary_rates(self, offset, length):
+                init = offset
                 string = ""
-                for byteRaw in rawdata:
-                    (byte,) = struct.unpack("B", bytearray([byteRaw])[0:1])
+                while (offset - init) < length:
+                    byte, = struct.unpack_from('B', self.data, offset)
                     r = byte & 0x7f
-
                     if r == BSS_MEMBERSHIP_SELECTOR_VHT_PHY and byte & 0x80:
                         string += "VHT"
                     elif r == BSS_MEMBERSHIP_SELECTOR_HT_PHY and byte & 0x80:
                         string += "HT"
                     else:
                         string += "%d.%d" % (r / 2, 5 * (r & 1))
-
+                    offset += 1
                     string += "%s " % ("*" if byte & 0x80 else "")
-
                 return string
 
-            def binary_tim(self, data):
-                (count,) = struct.unpack("B", data[0:1])
-                (period,) = struct.unpack("B", data[1:2])
-                (bitmapc,) = struct.unpack("B", data[2:3])
-                (bitmap0,) = struct.unpack("B", data[3:4])
+            def binary_tim(self, offset):
+                (count,
+                 period,
+                 bitmapc,
+                 bitmap0) = struct.unpack_from('BBBB',
+                                               self.data,
+                                               offset)
                 return ("DTIM Count {0} DTIM Period {1} Bitmap Control 0x{2} "
                         "Bitmap[0] 0x{3}".format(count,
                                                  period,
                                                  bitmapc,
                                                  bitmap0))
 
-            def binary_vendor(self, rawdata):
-                '''
-                Extract vendor data
-                '''
-                vendor = {}
-# pdb.set_trace()
-                size = len(rawdata)
-                # if len > 4 and rawdata[0] == ms_oui[0]
-                # and rawdata[1] == ms_oui[1] and rawdata[2] == ms_oui[2]
-
-                if size < 3:
-                    vendor["VENDOR_NAME"] = "Vendor specific: <too short data:"
-                    + hexdump(rawdata)
-                    return vendor
-
             def decode_nlas(self):
-
                 return
 
             def decode(self):
@@ -458,34 +468,52 @@ class nl80211cmd(genlmsg):
 
                 self.value = {}
 
-                init = self.buf.tell()
+                init = offset = self.offset + 4
 
-                while (self.buf.tell()-init) < (self.length-4):
-                    (msg_type, length) = struct.unpack('BB', self.buf.read(2))
-                    data = self.buf.read(length)
+                while (offset - init) < (self.length - 4):
+                    (msg_type, length) = struct.unpack_from('BB',
+                                                            self.data,
+                                                            offset)
                     if msg_type == NL80211_BSS_ELEMENTS_SSID:
-                        self.value["SSID"] = data
+                        self.value["SSID"], = (struct
+                                               .unpack_from('%is' % length,
+                                                            self.data,
+                                                            offset + 2))
 
                     if msg_type == NL80211_BSS_ELEMENTS_SUPPORTED_RATES:
-                        supported_rates = self.binary_rates(data)
+                        supported_rates = self.binary_rates(offset + 2, length)
                         self.value["SUPPORTED_RATES"] = supported_rates
 
                     if msg_type == NL80211_BSS_ELEMENTS_CHANNEL:
-                        (channel,) = struct.unpack("B", data[0:1])
+                        channel, = struct.unpack_from('B',
+                                                      self.data,
+                                                      offset + 2)
                         self.value["CHANNEL"] = channel
 
                     if msg_type == NL80211_BSS_ELEMENTS_TIM:
                         self.value["TRAFFIC INDICATION MAP"] = \
-                            self.binary_tim(data)
+                            self.binary_tim(offset + 2)
+
+                    if msg_type == NL80211_BSS_ELEMENTS_RSN:
+                        self.value["RSN"], = (struct
+                                              .unpack_from('%is' % length,
+                                                           self.data,
+                                                           offset + 2))
 
                     if msg_type == NL80211_BSS_ELEMENTS_EXTENDED_RATE:
-                        extended_rates = self.binary_rates(data)
+                        extended_rates = self.binary_rates(offset + 2, length)
                         self.value["EXTENDED_RATES"] = extended_rates
 
                     if msg_type == NL80211_BSS_ELEMENTS_VENDOR:
-                        self.binary_vendor(data)
+                        # There may be multiple vendor IEs, create a list
+                        if "VENDOR" not in self.value.keys():
+                            self.value["VENDOR"] = []
+                        vendor_ie, = (struct.unpack_from('%is' % length,
+                                                         self.data,
+                                                         offset + 2))
+                        self.value["VENDOR"].append(vendor_ie)
 
-                self.buf.seek(init)
+                    offset += length + 2
 
         prefix = 'NL80211_BSS_'
         nla_map = (('__NL80211_BSS_INVALID', 'hex'),
@@ -506,6 +534,105 @@ class nl80211cmd(genlmsg):
                    ('NL80211_BSS_MAX', 'hex')
                    )
 
+    class STAInfo(nla):
+        class STAFlags(nla_base):
+            '''
+            Decode the flags that may be set.
+            See nl80211.h: struct nl80211_sta_flag_update,
+            NL80211_STA_INFO_STA_FLAGS
+            '''
+
+            def decode_nlas(self):
+                return
+
+            def decode(self):
+                nla_base.decode(self)
+                self.value = {}
+                self.value["AUTHORIZED"] = False
+                self.value["SHORT_PREAMBLE"] = False
+                self.value["WME"] = False
+                self.value["MFP"] = False
+                self.value["AUTHENTICATED"] = False
+                self.value["TDLS_PEER"] = False
+                self.value["ASSOCIATED"] = False
+
+                init = offset = self.offset + 4
+                while (offset - init) < (self.length - 4):
+                    (msg_type, length) = struct.unpack_from('BB',
+                                                            self.data,
+                                                            offset)
+                    mask, set_ = struct.unpack_from('II',
+                                                    self.data,
+                                                    offset + 2)
+
+                    if mask & NL80211_STA_FLAG_AUTHORIZED:
+                        if set_ & NL80211_STA_FLAG_AUTHORIZED:
+                            self.value["AUTHORIZED"] = True
+
+                    if mask & NL80211_STA_FLAG_SHORT_PREAMBLE:
+                        if set_ & NL80211_STA_FLAG_SHORT_PREAMBLE:
+                            self.value["SHORT_PREAMBLE"] = True
+
+                    if mask & NL80211_STA_FLAG_WME:
+                        if set_ & NL80211_STA_FLAG_WME:
+                            self.value["WME"] = True
+
+                    if mask & NL80211_STA_FLAG_MFP:
+                        if set_ & NL80211_STA_FLAG_MFP:
+                            self.value["MFP"] = True
+
+                    if mask & NL80211_STA_FLAG_AUTHENTICATED:
+                        if set_ & NL80211_STA_FLAG_AUTHENTICATED:
+                            self.value["AUTHENTICATED"] = True
+
+                    if mask & NL80211_STA_FLAG_TDLS_PEER:
+                        if set_ & NL80211_STA_FLAG_TDLS_PEER:
+                            self.value["TDLS_PEER"] = True
+
+                    if mask & NL80211_STA_FLAG_ASSOCIATED:
+                        if set_ & NL80211_STA_FLAG_ASSOCIATED:
+                            self.value["ASSOCIATED"] = True
+
+                    offset += length + 2
+
+        prefix = 'NL80211_STA_INFO_'
+        nla_map = (('__NL80211_STA_INFO_INVALID', 'hex'),
+                   ('NL80211_STA_INFO_INACTIVE_TIME', 'uint32'),
+                   ('NL80211_STA_INFO_RX_BYTES', 'uint32'),
+                   ('NL80211_STA_INFO_TX_BYTES', 'uint32'),
+                   ('NL80211_STA_INFO_LLID', 'uint16'),
+                   ('NL80211_STA_INFO_PLID', 'uint16'),
+                   ('NL80211_STA_INFO_PLINK_STATE', 'uint8'),
+                   ('NL80211_STA_INFO_SIGNAL', 'int8'),
+                   ('NL80211_STA_INFO_TX_BITRATE', 'hex'),
+                   ('NL80211_STA_INFO_RX_PACKETS', 'uint32'),
+                   ('NL80211_STA_INFO_TX_PACKETS', 'uint32'),
+                   ('NL80211_STA_INFO_TX_RETRIES', 'uint32'),
+                   ('NL80211_STA_INFO_TX_FAILED', 'uint32'),
+                   ('NL80211_STA_INFO_SIGNAL_AVG', 'int8'),
+                   ('NL80211_STA_INFO_RX_BITRATE', 'hex'),
+                   ('NL80211_STA_INFO_BSS_PARAM', 'hex'),
+                   ('NL80211_STA_INFO_CONNECTED_TIME', 'uint32'),
+                   ('NL80211_STA_INFO_STA_FLAGS', 'STAFlags'),
+                   ('NL80211_STA_INFO_BEACON_LOSS', 'uint32'),
+                   ('NL80211_STA_INFO_T_OFFSET', 'int64'),
+                   ('NL80211_STA_INFO_LOCAL_PM', 'hex'),
+                   ('NL80211_STA_INFO_PEER_PM', 'hex'),
+                   ('NL80211_STA_INFO_NONPEER_PM', 'hex'),
+                   ('NL80211_STA_INFO_RX_BYTES64', 'uint64'),
+                   ('NL80211_STA_INFO_TX_BYTES64', 'uint64'),
+                   ('NL80211_STA_INFO_CHAIN_SIGNAL', 'string'),
+                   ('NL80211_STA_INFO_CHAIN_SIGNAL_AVG', 'string'),
+                   ('NL80211_STA_INFO_EXPECTED_THROUGHPUT', 'uint32'),
+                   ('NL80211_STA_INFO_RX_DROP_MISC', 'uint32'),
+                   ('NL80211_STA_INFO_BEACON_RX', 'uint64'),
+                   ('NL80211_STA_INFO_BEACON_SIGNAL_AVG', 'uint8'),
+                   ('NL80211_STA_INFO_TID_STATS', 'hex'),
+                   ('NL80211_STA_INFO_RX_DURATION', 'uint64'),
+                   ('NL80211_STA_INFO_PAD', 'hex'),
+                   ('NL80211_STA_INFO_MAX', 'hex')
+                   )
+
 
 class MarshalNl80211(Marshal):
     msg_map = {NL80211_CMD_UNSPEC: nl80211cmd,
@@ -641,6 +768,6 @@ class NL80211(GenericNetlinkSocket):
         GenericNetlinkSocket.__init__(self)
         self.marshal = MarshalNl80211()
 
-    def bind(self, groups=0, async=False):
+    def bind(self, groups=0, **kwarg):
         GenericNetlinkSocket.bind(self, 'nl80211', nl80211cmd,
-                                  groups, None, async)
+                                  groups, None, **kwarg)
