@@ -11,8 +11,8 @@ asynchronous I/O
 ----------------
 
 To run async reader thread, one should call
-`NetlinkSocket.bind(async=True)`. In that case a
-background thread will be launched. The thread will
+`NetlinkSocket.bind(async_cache=True)`. In that case
+a background thread will be launched. The thread will
 automatically collect all the messages and store
 into a userspace buffer.
 
@@ -103,6 +103,8 @@ from pyroute2.netlink import nlmsg
 from pyroute2.netlink import mtypes
 from pyroute2.netlink import NLMSG_ERROR
 from pyroute2.netlink import NLMSG_DONE
+from pyroute2.netlink import NETLINK_ADD_MEMBERSHIP
+from pyroute2.netlink import NETLINK_DROP_MEMBERSHIP
 from pyroute2.netlink import NETLINK_GENERIC
 from pyroute2.netlink import NETLINK_LISTEN_ALL_NSID
 from pyroute2.netlink import NLM_F_DUMP
@@ -129,6 +131,7 @@ class Marshal(object):
     msg_map = {}
     type_offset = 4
     type_format = 'H'
+    error_type = NLMSG_ERROR
     debug = False
 
     def __init__(self):
@@ -159,8 +162,8 @@ class Marshal(object):
             msg_type, = struct.unpack_from(self.type_format,
                                            data,
                                            offset + self.type_offset)
-            if msg_type == NLMSG_ERROR:
-                code = abs(struct.unpack_from('i', data, offset+16)[0])
+            if msg_type == self.error_type:
+                code = abs(struct.unpack_from('i', data, offset + 16)[0])
                 if code > 0:
                     error = NetlinkError(code)
 
@@ -172,9 +175,9 @@ class Marshal(object):
                 msg['header']['error'] = error
                 # try to decode encapsulated error message
                 if error is not None:
-                    enc_type = struct.unpack_from('H', data, offset+24)[0]
+                    enc_type = struct.unpack_from('H', data, offset + 24)[0]
                     enc_class = self.msg_map.get(enc_type, nlmsg)
-                    enc = enc_class(data, offset=offset+20)
+                    enc = enc_class(data, offset=offset + 20)
                     enc.decode()
                     msg['header']['errmsg'] = enc
                 if callback and seq == msg['header']['sequence_number']:
@@ -281,6 +284,8 @@ class NetlinkMixin(object):
                  port=None,
                  pid=None,
                  fileno=None,
+                 sndbuf=1048576,
+                 rcvbuf=1048576,
                  all_ns=False):
         #
         # That's a trick. Python 2 is not able to construct
@@ -305,6 +310,8 @@ class NetlinkMixin(object):
         self.fixed = True
         self.family = family
         self._fileno = fileno
+        self._sndbuf = sndbuf
+        self._rcvbuf = rcvbuf
         self.backlog = {0: []}
         self.callbacks = []     # [(predicate, callback, args), ...]
         self.pthread = None
@@ -315,6 +322,7 @@ class NetlinkMixin(object):
                              'provide_master': config.kernel[0] > 2}
         self.backlog_lock = threading.Lock()
         self.read_lock = threading.Lock()
+        self.sys_lock = threading.Lock()
         self.change_master = threading.Event()
         self.lock = LockFactory()
         self._sock = None
@@ -337,6 +345,20 @@ class NetlinkMixin(object):
         self.groups = 0
         self.marshal = Marshal()
         # 8<-----------------------------------------
+        if not config.nlm_generator:
+
+            def nlm_request(*argv, **kwarg):
+                return tuple(self._genlm_request(*argv, **kwarg))
+
+            def get(*argv, **kwarg):
+                return tuple(self._genlm_get(*argv, **kwarg))
+
+            self._genlm_request = self.nlm_request
+            self._genlm_get = self.get
+
+            self.nlm_request = nlm_request
+            self.get = get
+
         # Set defaults
         self.post_init()
 
@@ -344,6 +366,9 @@ class NetlinkMixin(object):
         return type(self)(family=self.family)
 
     def close(self):
+        if self.pthread:
+            self.buffer_queue.put(struct.pack('IHHQIQQ',
+                                              28, 2, 0, 0, 104, 0, 0))
         try:
             os.close(self._ctrl_write)
             os.close(self._ctrl_read)
@@ -596,218 +621,233 @@ class NetlinkMixin(object):
                 # get bufsize from SO_RCVBUF
                 bufsize = self.getsockopt(SOL_SOCKET, SO_RCVBUF) // 2
 
-            ret = []
+            tmsg = None
             enough = False
-            while not enough:
-                # 8<-----------------------------------------------------------
-                #
-                # This stage changes the backlog, so use mutex to
-                # prevent side changes
-                self.backlog_lock.acquire()
-                ##
-                # Stage 1. BEGIN
-                #
-                # 8<-----------------------------------------------------------
-                #
-                # Check backlog and return already collected
-                # messages.
-                #
-                if msg_seq == 0 and self.backlog[0]:
-                    # Zero queue.
+            backlog_acquired = False
+            try:
+                while not enough:
+                    # 8<-----------------------------------------------------------
                     #
-                    # Load the backlog, if there is valid
-                    # content in it
-                    ret.extend(self.backlog[0])
-                    self.backlog[0] = []
-                    # And just exit
-                    self.backlog_lock.release()
-                    break
-                elif self.backlog.get(msg_seq, None):
-                    # Any other msg_seq.
+                    # This stage changes the backlog, so use mutex to
+                    # prevent side changes
+                    self.backlog_lock.acquire()
+                    backlog_acquired = True
+                    ##
+                    # Stage 1. BEGIN
                     #
-                    # Collect messages up to the terminator.
-                    # Terminator conditions:
-                    #  * NLMSG_ERROR != 0
-                    #  * NLMSG_DONE
-                    #  * terminate() function (if defined)
-                    #  * not NLM_F_MULTI
+                    # 8<-----------------------------------------------------------
                     #
-                    # Please note, that if terminator not occured,
-                    # more `recv()` rounds CAN be required.
-                    for msg in tuple(self.backlog[msg_seq]):
+                    # Check backlog and return already collected
+                    # messages.
+                    #
+                    if msg_seq == 0 and self.backlog[0]:
+                        # Zero queue.
+                        #
+                        # Load the backlog, if there is valid
+                        # content in it
+                        for msg in self.backlog[0]:
+                            yield msg
+                        self.backlog[0] = []
+                        # And just exit
+                        break
+                    elif msg_seq != 0 and len(self.backlog.get(msg_seq, [])):
+                        # Any other msg_seq.
+                        #
+                        # Collect messages up to the terminator.
+                        # Terminator conditions:
+                        #  * NLMSG_ERROR != 0
+                        #  * NLMSG_DONE
+                        #  * terminate() function (if defined)
+                        #  * not NLM_F_MULTI
+                        #
+                        # Please note, that if terminator not occured,
+                        # more `recv()` rounds CAN be required.
+                        for msg in tuple(self.backlog[msg_seq]):
 
-                        # Drop the message from the backlog, if any
-                        self.backlog[msg_seq].remove(msg)
+                            # Drop the message from the backlog, if any
+                            self.backlog[msg_seq].remove(msg)
 
-                        # If there is an error, raise exception
-                        if msg['header'].get('error', None) is not None:
-                            self.backlog[0].extend(self.backlog[msg_seq])
-                            del self.backlog[msg_seq]
-                            # The loop is done
-                            self.backlog_lock.release()
-                            raise msg['header']['error']
+                            # If there is an error, raise exception
+                            if msg['header'].get('error', None) is not None:
+                                self.backlog[0].extend(self.backlog[msg_seq])
+                                del self.backlog[msg_seq]
+                                # The loop is done
+                                raise msg['header']['error']
 
-                        # If it is the terminator message, say "enough"
-                        # and requeue all the rest into Zero queue
-                        if (msg['header']['type'] == NLMSG_DONE) or \
-                                (terminate is not None and terminate(msg)):
-                            # The loop is done
-                            enough = True
-
-                        # If it is just a normal message, append it to
-                        # the response
-                        if not enough:
-                            ret.append(msg)
-                            # But finish the loop on single messages
-                            if not msg['header']['flags'] & NLM_F_MULTI:
-                                # but not multi -- so end the loop
+                            # If it is the terminator message, say "enough"
+                            # and requeue all the rest into Zero queue
+                            if terminate is not None:
+                                tmsg = terminate(msg)
+                                if isinstance(tmsg, nlmsg):
+                                    yield msg
+                            if (msg['header']['type'] == NLMSG_DONE) or tmsg:
+                                # The loop is done
                                 enough = True
 
-                        # Enough is enough, requeue the rest and delete
-                        # our backlog
-                        if enough:
+                            # If it is just a normal message, append it to
+                            # the response
+                            if not enough:
+                                # finish the loop on single messages
+                                if not msg['header']['flags'] & NLM_F_MULTI:
+                                    enough = True
+                                yield msg
+
+                            # Enough is enough, requeue the rest and delete
+                            # our backlog
+                            if enough:
+                                self.backlog[0].extend(self.backlog[msg_seq])
+                                del self.backlog[msg_seq]
+                                break
+
+                        # Next iteration
+                        self.backlog_lock.release()
+                        backlog_acquired = False
+                    else:
+                        # Stage 1. END
+                        #
+                        # 8<-------------------------------------------------------
+                        #
+                        # Stage 2. BEGIN
+                        #
+                        # 8<-------------------------------------------------------
+                        #
+                        # Receive the data from the socket and put the messages
+                        # into the backlog
+                        #
+                        self.backlog_lock.release()
+                        backlog_acquired = False
+                        ##
+                        #
+                        # Control the timeout. We should not be within the
+                        # function more than TIMEOUT seconds. All the locks
+                        # MUST be released here.
+                        #
+                        if (msg_seq != 0) and \
+                                (time.time() - ctime > self.get_timeout):
+                            # requeue already received for that msg_seq
                             self.backlog[0].extend(self.backlog[msg_seq])
                             del self.backlog[msg_seq]
-                            break
-
-                    # Next iteration
-                    self.backlog_lock.release()
-                else:
-                    # Stage 1. END
-                    #
-                    # 8<-------------------------------------------------------
-                    #
-                    # Stage 2. BEGIN
-                    #
-                    # 8<-------------------------------------------------------
-                    #
-                    # Receive the data from the socket and put the messages
-                    # into the backlog
-                    #
-                    self.backlog_lock.release()
-                    ##
-                    #
-                    # Control the timeout. We should not be within the
-                    # function more than TIMEOUT seconds. All the locks
-                    # MUST be released here.
-                    #
-                    if time.time() - ctime > self.get_timeout:
-                        if self.get_timeout_exception:
-                            raise self.get_timeout_exception()
-                        else:
-                            return ret
-                    #
-                    if self.read_lock.acquire(False):
-                        self.change_master.clear()
-                        # If the socket is free to read from, occupy
-                        # it and wait for the data
-                        #
-                        # This is a time consuming process, so all the
-                        # locks, except the read lock must be released
-                        data = self.recv_ft(bufsize)
-                        # Parse data
-                        msgs = self.marshal.parse(data, msg_seq, callback)
-                        # Reset ctime -- timeout should be measured
-                        # for every turn separately
-                        ctime = time.time()
-                        #
-                        current = self.buffer_queue.qsize()
-                        delta = current - self.qsize
-                        if delta > 10:
-                            delay = min(3, max(0.01, float(current) / 60000))
-                            message = ("Packet burst: "
-                                       "delta=%s qsize=%s delay=%s"
-                                       % (delta, current, delay))
-                            if delay < 1:
-                                log.debug(message)
+                            # throw an exception
+                            if self.get_timeout_exception:
+                                raise self.get_timeout_exception()
                             else:
-                                log.warning(message)
-                            time.sleep(delay)
-                        self.qsize = current
+                                return
+                        #
+                        if self.read_lock.acquire(False):
+                            try:
+                                self.change_master.clear()
+                                # If the socket is free to read from, occupy
+                                # it and wait for the data
+                                #
+                                # This is a time consuming process, so all the
+                                # locks, except the read lock must be released
+                                data = self.recv_ft(bufsize)
+                                # Parse data
+                                msgs = self.marshal.parse(data,
+                                                          msg_seq,
+                                                          callback)
+                                # Reset ctime -- timeout should be measured
+                                # for every turn separately
+                                ctime = time.time()
+                                #
+                                current = self.buffer_queue.qsize()
+                                delta = current - self.qsize
+                                if delta > 10:
+                                    delay = min(3, max(0.01,
+                                                       float(current) / 60000))
+                                    message = ("Packet burst: "
+                                               "delta=%s qsize=%s delay=%s"
+                                               % (delta, current, delay))
+                                    if delay < 1:
+                                        log.debug(message)
+                                    else:
+                                        log.warning(message)
+                                    time.sleep(delay)
+                                self.qsize = current
 
-                        # We've got the data, lock the backlog again
-                        self.backlog_lock.acquire()
-                        for msg in msgs:
-                            seq = msg['header']['sequence_number']
-                            if seq not in self.backlog:
-                                if msg['header']['type'] == NLMSG_ERROR:
-                                    # Drop orphaned NLMSG_ERROR messages
-                                    continue
-                                seq = 0
-                            # 8<-----------------------------------------------
-                            # Callbacks section
-                            for cr in self.callbacks:
-                                try:
-                                    if cr[0](msg):
-                                        cr[1](msg, *cr[2])
-                                except:
-                                    log.warning("Callback fail: %s" % (cr))
-                                    log.warning(traceback.format_exc())
-                            # 8<-----------------------------------------------
-                            self.backlog[seq].append(msg)
-                        # We finished with the backlog, so release the lock
-                        self.backlog_lock.release()
+                                # We've got the data, lock the backlog again
+                                with self.backlog_lock:
+                                    for msg in msgs:
+                                        seq = msg['header']['sequence_number']
+                                        if seq not in self.backlog:
+                                            if msg['header']['type'] == \
+                                                    NLMSG_ERROR:
+                                                # Drop orphaned NLMSG_ERROR
+                                                # messages
+                                                continue
+                                            seq = 0
+                                        # 8<-----------------------------------
+                                        # Callbacks section
+                                        for cr in self.callbacks:
+                                            try:
+                                                if cr[0](msg):
+                                                    cr[1](msg, *cr[2])
+                                            except:
+                                                # FIXME
+                                                #
+                                                # Usually such code formatting
+                                                # means that the method should
+                                                # be refactored to avoid such
+                                                # indentation.
+                                                #
+                                                # Plz do something with it.
+                                                #
+                                                lw = log.warning
+                                                lw("Callback fail: %s" % (cr))
+                                                lw(traceback.format_exc())
+                                        # 8<-----------------------------------
+                                        self.backlog[seq].append(msg)
 
-                        # Now wake up other threads
-                        self.change_master.set()
-
-                        # Finally, release the read lock: all data processed
-                        self.read_lock.release()
-                    else:
-                        # If the socket is occupied and there is still no
-                        # data for us, wait for the next master change or
-                        # for a timeout
-                        self.change_master.wait(1)
-                    # 8<-------------------------------------------------------
-                    #
-                    # Stage 2. END
-                    #
-                    # 8<-------------------------------------------------------
-
-            return ret
+                                # Now wake up other threads
+                                self.change_master.set()
+                            finally:
+                                # Finally, release the read lock: all data
+                                # processed
+                                self.read_lock.release()
+                        else:
+                            # If the socket is occupied and there is still no
+                            # data for us, wait for the next master change or
+                            # for a timeout
+                            self.change_master.wait(1)
+                        # 8<-------------------------------------------------------
+                        #
+                        # Stage 2. END
+                        #
+                        # 8<-------------------------------------------------------
+            finally:
+                if backlog_acquired:
+                    self.backlog_lock.release()
 
     def nlm_request(self, msg, msg_type,
                     msg_flags=NLM_F_REQUEST | NLM_F_DUMP,
                     terminate=None,
-                    callback=None,
-                    exception_catch=Exception,
-                    exception_handler=None):
+                    callback=None):
 
-        def do_try():
-            msg_seq = self.addr_pool.alloc()
-            with self.lock[msg_seq]:
-                try:
-                    self.put(msg, msg_type, msg_flags, msg_seq=msg_seq)
-                    ret = self.get(msg_seq=msg_seq,
-                                   terminate=terminate,
-                                   callback=callback)
-                    return ret
-                except Exception:
-                    raise
-                finally:
-                    # Ban this msg_seq for 0xff rounds
-                    #
-                    # It's a long story. Modern kernels for RTM_SET.*
-                    # operations always return NLMSG_ERROR(0) == success,
-                    # even not setting NLM_F_MULTY flag on other response
-                    # messages and thus w/o any NLMSG_DONE. So, how to detect
-                    # the response end? One can not rely on NLMSG_ERROR on
-                    # old kernels, but we have to support them too. Ty, we
-                    # just ban msg_seq for several rounds, and NLMSG_ERROR,
-                    # being received, will become orphaned and just dropped.
-                    #
-                    # Hack, but true.
-                    self.addr_pool.free(msg_seq, ban=0xff)
-
-        while True:
+        msg_seq = self.addr_pool.alloc()
+        with self.lock[msg_seq]:
             try:
-                return do_try()
-            except exception_catch as e:
-                if exception_handler and not exception_handler(e):
-                    continue
-                raise
+                self.put(msg, msg_type, msg_flags, msg_seq=msg_seq)
+                for msg in self.get(msg_seq=msg_seq,
+                                    terminate=terminate,
+                                    callback=callback):
+                    yield msg
+
             except Exception:
                 raise
+            finally:
+                # Ban this msg_seq for 0xff rounds
+                #
+                # It's a long story. Modern kernels for RTM_SET.*
+                # operations always return NLMSG_ERROR(0) == success,
+                # even not setting NLM_F_MULTY flag on other response
+                # messages and thus w/o any NLMSG_DONE. So, how to detect
+                # the response end? One can not rely on NLMSG_ERROR on
+                # old kernels, but we have to support them too. Ty, we
+                # just ban msg_seq for several rounds, and NLMSG_ERROR,
+                # being received, will become orphaned and just dropped.
+                #
+                # Hack, but true.
+                self.addr_pool.free(msg_seq, ban=0xff)
 
 
 class BatchAddrPool(object):
@@ -865,24 +905,13 @@ class NetlinkSocket(NetlinkMixin):
 
     def post_init(self):
         # recreate the underlying socket
-        with self.lock:
+        with self.sys_lock:
             if self._sock is not None:
                 self._sock.close()
             self._sock = config.SocketBase(AF_NETLINK,
                                            SOCK_DGRAM,
                                            self.family,
                                            self._fileno)
-            for name in ('getsockname', 'getsockopt', 'makefile',
-                         'setsockopt', 'setblocking', 'settimeout',
-                         'gettimeout', 'shutdown', 'recvfrom',
-                         'recvfrom_into', 'fileno'):
-                setattr(self, name, getattr(self._sock, name))
-
-            self._sendto = getattr(self._sock, 'sendto')
-            self._recv = getattr(self._sock, 'recv')
-            self._recv_into = getattr(self._sock, 'recv_into')
-            # setup fast-track
-            self.recv_ft = getattr(self._sock, 'recv')
             self.sendto_gate = self._gate
 
             # monkey patch recv_into on Python 2.6
@@ -893,17 +922,30 @@ class NetlinkSocket(NetlinkMixin):
                 def patch(data, bsize):
                     data[0:] = self._sock.recv(bsize)
                 self._sock.recv_into = patch
-            self.setsockopt(SOL_SOCKET, SO_SNDBUF, 1024 * 1024)
-            self.setsockopt(SOL_SOCKET, SO_RCVBUF, 1024 * 1024)
+            self.setsockopt(SOL_SOCKET, SO_SNDBUF, self._sndbuf)
+            self.setsockopt(SOL_SOCKET, SO_RCVBUF, self._rcvbuf)
             if self.all_ns:
                 self.setsockopt(SOL_NETLINK, NETLINK_LISTEN_ALL_NSID, 1)
 
+    def __getattr__(self, attr):
+        if attr in ('getsockname', 'getsockopt', 'makefile',
+                    'setsockopt', 'setblocking', 'settimeout',
+                    'gettimeout', 'shutdown', 'recvfrom',
+                    'recvfrom_into', 'fileno'):
+            return getattr(self._sock, attr)
+        elif attr in ('_sendto', '_recv', '_recv_into'):
+            return getattr(self._sock, attr.lstrip("_"))
+        elif attr == "recv_ft":
+            return self._sock.recv
+
+        raise AttributeError(attr)
+
     def _gate(self, msg, addr):
         msg.reset()
         msg.encode()
         return self._sock.sendto(msg.data, addr)
 
-    def bind(self, groups=0, pid=None, async=False):
+    def bind(self, groups=0, pid=None, **kwarg):
         '''
         Bind the socket to given multicast groups, using
         given pid.
@@ -917,6 +959,14 @@ class NetlinkSocket(NetlinkMixin):
             self.fixed = True
             self.pid = pid or os.getpid()
 
+        if 'async' in kwarg:
+            # FIXME
+            # raise deprecation error after 0.5.3
+            #
+            log.warning('use "async_cache" instead of "async", '
+                        '"async" is a keyword from Python 3.7')
+        async_cache = kwarg.get('async_cache') or kwarg.get('async')
+
         self.groups = groups
         # if we have pre-defined port, use it strictly
         if self.fixed:
@@ -936,7 +986,7 @@ class NetlinkSocket(NetlinkMixin):
             else:
                 raise KeyError('no free address available')
         # all is OK till now, so start async recv, if we need
-        if async:
+        if async_cache:
             def recv_plugin(*argv, **kwarg):
                 data_in = self.buffer_queue.get()
                 if isinstance(data_in, Exception):
@@ -959,11 +1009,17 @@ class NetlinkSocket(NetlinkMixin):
             self.pthread.setDaemon(True)
             self.pthread.start()
 
+    def add_membership(self, group):
+        self.setsockopt(SOL_NETLINK, NETLINK_ADD_MEMBERSHIP, group)
+
+    def drop_membership(self, group):
+        self.setsockopt(SOL_NETLINK, NETLINK_DROP_MEMBERSHIP, group)
+
     def close(self):
         '''
         Correctly close the socket and free all resources.
         '''
-        with self.lock:
+        with self.sys_lock:
             if self.closed:
                 return
             self.closed = True
