@@ -300,6 +300,28 @@ class ifinfbase(object):
     '''
     prefix = 'IFLA_'
 
+    #
+    # Changed from PRIMARY KEY to NOT NULL to support multiple
+    # targets in one table, so we can collect info from multiple
+    # systems.
+    #
+    # To provide data integrity one should use foreign keys,
+    # but when you create a foreign key using interfaces as
+    # the parent table, create also a unique index on
+    # the fields specified in the foreign key definition.
+    #
+    # E.g.
+    #
+    # CREATE TABLE interfaces (f_target TEXT NOT NULL,
+    #                          f_index INTEGER NOT NULL, ...)
+    # CREATE TABLE routes (f_target TEXT NOT NULL,
+    #                      f_RTA_OIF INTEGER, ...
+    #                      FOREIGN KEY (f_target, f_RTA_OIF)
+    #                      REFERENCES interfaces(f_target, f_index))
+    # CREATE UNIQUE INDEX if_idx ON interfaces(f_target, f_index)
+    #
+    sql_constraints = {'index': 'NOT NULL'}
+
     fields = (('family', 'B'),
               ('__align', 'x'),
               ('ifi_type', 'H'),
@@ -348,7 +370,15 @@ class ifinfbase(object):
                ('IFLA_PHYS_PORT_NAME', 'asciiz'),
                ('IFLA_PROTO_DOWN', 'uint8'),
                ('IFLA_GSO_MAX_SEGS', 'uint32'),
-               ('IFLA_GSO_MAX_SIZE', 'uint32'))
+               ('IFLA_GSO_MAX_SIZE', 'uint32'),
+               ('IFLA_PAD', 'hex'),
+               ('IFLA_XDP', 'hex'),
+               ('IFLA_EVENT', 'hex'),
+               ('IFLA_NEW_NETNSID', 'hex'),
+               ('IFLA_IF_NETNSID', 'hex'),
+               ('IFLA_CARRIER_UP_COUNT', 'hex'),
+               ('IFLA_CARRIER_DOWN_COUNT', 'hex'),
+               ('IFLA_NEW_IFINDEX', 'hex'))
 
     @staticmethod
     def flags2names(flags, mask=0xffffffff):
@@ -739,7 +769,7 @@ class ifinfbase(object):
                                'router_solicitation_delay',
                                'use_tempaddr',
                                'temp_valid_lft',
-                               'temp_prefered_lft',
+                               'temp_preferred_lft',
                                'regen_max_retry',
                                'max_desync_factor',
                                'max_addresses',
