@@ -15,7 +15,9 @@ from pyroute2.netlink.rtnl.fibmsg import FR_ACT_NAMES
 
 
 encap_types = {'mpls': 1,
-               AF_MPLS: 1}
+               AF_MPLS: 1,
+               'seg6': 5,
+               'bpf': 6}
 
 
 class IPRequest(dict):
@@ -79,6 +81,10 @@ class IPRouteRequest(IPRequest):
                'proto': rt_proto,
                'scope': rt_scope}
 
+    def __init__(self, obj=None):
+        self._mask = []
+        IPRequest.__init__(self, obj)
+
     def encap_header(self, header):
         '''
         Encap header transform. Format samples:
@@ -119,6 +125,76 @@ class IPRouteRequest(IPRequest):
             if override_bos:
                 ret[-1]['bos'] = 1
             return {'attrs': [['MPLS_IPTUNNEL_DST', ret]]}
+        '''
+        Seg6 encap header transform. Format samples:
+
+            {'type': 'seg6',
+             'mode': 'encap',
+             'segs': '2000::5,2000::6'}
+
+            {'type': 'seg6',
+             'mode': 'encap'
+             'segs': '2000::5,2000::6',
+             'hmac': 1}
+        '''
+        if header['type'] == 'seg6':
+            # Init step
+            ret = {}
+            # Parse segs
+            segs = header['segs']
+            # If they are in the form in_addr6,in_addr6
+            if isinstance(segs, basestring):
+                # Create an array with the splitted values
+                temp = segs.split(',')
+                # Init segs
+                segs = []
+                # Iterate over the values
+                for seg in temp:
+                    # Discard empty string
+                    if seg != '':
+                        # Add seg to segs
+                        segs.append(seg)
+            # Retrieve mode
+            mode = header['mode']
+            # hmac is optional and contains the hmac key
+            hmac = header.get('hmac', None)
+            # Construct the new object
+            ret = {'mode': mode, 'segs': segs}
+            # If hmac is present convert to u32
+            if hmac:
+                # Add to ret the hmac key
+                ret['hmac'] = hmac & 0xffffffff
+            # Done return the object
+            return {'attrs': [['SEG6_IPTUNNEL_SRH', ret]]}
+        '''
+        BPF encap header transform. Format samples:
+
+            {'type': 'bpf',
+             'in': {'fd':4, 'name':'firewall'}}
+
+            {'type': 'bpf',
+             'in'  : {'fd':4, 'name':'firewall'},
+             'out' : {'fd':5, 'name':'stats'},
+             'xmit': {'fd':6, 'name':'vlan_push', 'headroom':4}}
+        '''
+        if header['type'] == 'bpf':
+            attrs = {}
+            for key, value in header.items():
+                if key not in ['in', 'out', 'xmit']:
+                    continue
+
+                obj = [['LWT_BPF_PROG_FD', value['fd']],
+                       ['LWT_BPF_PROG_NAME', value['name']]]
+                if key == 'in':
+                    attrs['LWT_BPF_IN'] = {'attrs': obj}
+                elif key == 'out':
+                    attrs['LWT_BPF_OUT'] = {'attrs': obj}
+                elif key == 'xmit':
+                    attrs['LWT_BPF_XMIT'] = {'attrs': obj}
+                    if 'headroom' in value:
+                        attrs['LWT_BPF_XMIT_HEADROOM'] = value['headroom']
+
+            return {'attrs': attrs.items()}
 
     def mpls_rta(self, value):
         ret = []
@@ -153,25 +229,31 @@ class IPRouteRequest(IPRequest):
             dict.__setitem__(self, 'type', 1)
         elif key == 'flags' and self.get('family', None) == AF_MPLS:
             return
-        elif key == 'dst':
+        elif key in ('dst', 'src'):
             if isinstance(value, dict):
-                dict.__setitem__(self, 'dst', value)
+                dict.__setitem__(self, key, value)
             elif isinstance(value, int):
-                dict.__setitem__(self, 'dst', {'label': value,
-                                               'bos': 1})
+                dict.__setitem__(self, key, {'label': value,
+                                             'bos': 1})
             elif value != 'default':
                 value = value.split('/')
+                mask = None
                 if len(value) == 1:
                     dst = value[0]
-                    mask = 0
+                    if self.get('family', 0) == AF_INET:
+                        mask = 32
+                    elif self.get('family', 0) == AF_INET6:
+                        mask = 128
+                    else:
+                        self._mask.append('%s_len' % key)
                 elif len(value) == 2:
                     dst = value[0]
                     mask = int(value[1])
                 else:
-                    raise ValueError('wrong destination')
-                dict.__setitem__(self, 'dst', dst)
-                if mask:
-                    dict.__setitem__(self, 'dst_len', mask)
+                    raise ValueError('wrong address spec')
+                dict.__setitem__(self, key, dst)
+                if mask is not None:
+                    dict.__setitem__(self, '%s_len' % key, mask)
         elif key == 'newdst':
             dict.__setitem__(self, 'newdst', self.mpls_rta(value))
         elif key in self.resolve.keys():
@@ -192,6 +274,40 @@ class IPRouteRequest(IPRequest):
                                                      value['type']))
                     dict.__setitem__(self, 'encap',
                                      self.encap_header(value))
+                # human-friendly form:
+                #
+                # 'encap': {'type': 'seg6',
+                #           'mode': 'encap'
+                #           'segs': '2000::5,2000::6'}
+                #
+                # 'encap': {'type': 'seg6',
+                #           'mode': 'inline'
+                #           'segs': '2000::5,2000::6'
+                #           'hmac': 1}
+                #
+                # 'encap': {'type': 'seg6',
+                #           'mode': 'encap'
+                #           'segs': '2000::5,2000::6'
+                #           'hmac': 0xf}
+                #
+                # 'encap': {'type': 'seg6',
+                #           'mode': 'inline'
+                #           'segs': ['2000::5', '2000::6']}
+                #
+                # 'type', 'mode' and 'segs' are mandatory
+                if 'type' in value and 'mode' in value and 'segs' in value:
+                    dict.__setitem__(self, 'encap_type',
+                                     encap_types.get(value['type'],
+                                                     value['type']))
+                    dict.__setitem__(self, 'encap',
+                                     self.encap_header(value))
+                elif 'type' in value and ('in' in value or 'out' in value or
+                                          'xmit' in value):
+                    dict.__setitem__(self, 'encap_type',
+                                     encap_types.get(value['type'],
+                                                     value['type']))
+                    dict.__setitem__(self, 'encap',
+                                     self.encap_header(value))
                 # assume it is a ready-to-use NLA
                 elif 'attrs' in value:
                     dict.__setitem__(self, 'encap', value)
@@ -243,6 +359,14 @@ class IPRouteRequest(IPRequest):
                 ret.append(nh)
             if ret:
                 dict.__setitem__(self, 'multipath', ret)
+        elif key == 'family':
+            for d in self._mask:
+                if value == AF_INET:
+                    dict.__setitem__(self, d, 32)
+                elif value == AF_INET6:
+                    dict.__setitem__(self, d, 128)
+            self._mask = []
+            dict.__setitem__(self, key, value)
         else:
             dict.__setitem__(self, key, value)
 
