@@ -99,8 +99,10 @@ __NR = {'x86_': {'64bit': 308},
         'mips': {'32bit': 4344,
                  '64bit': 5303},  # FIXME: NABI32?
         'armv': {'32bit': 375},
-        'aarc': {'64bit': 268},  # FIXME: EABI vs. OABI?
-        'ppc6': {'64bit': 350}}
+        'aarc': {'32bit': 375,
+                 '64bit': 268},  # FIXME: EABI vs. OABI?
+        'ppc6': {'64bit': 350},
+        's390': {'64bit': 339}}
 __NR_setns = __NR.get(config.machine[:4], {}).get(config.arch, 308)
 
 CLONE_NEWNET = 0x40000000
@@ -218,7 +220,11 @@ def setns(netns, flags=os.O_CREAT, libc=None):
 
         - O_CREAT -- create netns, if doesn't exist
         - O_CREAT | O_EXCL -- create only if doesn't exist
+
+    Changed in 0.5.1: the routine closes the ns fd if it's
+    not provided via arguments.
     '''
+    newfd = False
     libc = libc or ctypes.CDLL('libc.so.6', use_errno=True)
     if isinstance(netns, basestring):
         netnspath = _get_netnspath(netns)
@@ -229,16 +235,15 @@ def setns(netns, flags=os.O_CREAT, libc=None):
             if flags & os.O_CREAT:
                 create(netns, libc=libc)
         nsfd = os.open(netnspath, os.O_RDONLY)
-        ret = nsfd
+        newfd = True
     elif isinstance(netns, file):
         nsfd = netns.fileno()
-        ret = netns
     elif isinstance(netns, int):
         nsfd = netns
-        ret = netns
     else:
         raise RuntimeError('netns should be a string or an open fd')
     error = libc.syscall(__NR_setns, nsfd, CLONE_NEWNET)
+    if newfd:
+        os.close(nsfd)
     if error != 0:
         raise OSError(ctypes.get_errno(), 'failed to open netns', netns)
-    return ret
